/*
sapsum: inspired by md5sum.

Written by Michael Paul Johnson in August 2000 and released to
the Public Domain. Derived from sha1sum.

This program computes Sapphire II Stream Cipher 32-byte hashes
of files, and optionally compares them with stored hashes.

*/

#include <stdio.h>
#include <string.h>
#ifdef __BORLANDC__
#include <dir.h>
#include <dos.h>
#include <process.h>   /*  prototype for exit() - JHB
               needed for Win32, but chokes Linux - MPJ */
#else
#include <unistd.h>
#include <stdlib.h>
#endif
#include <ctype.h>
#include "sapphire.h"

#define LINESIZE 2048
#define HASHSIZE 32

/* hashfile computes the hash of the named file and puts
   it in the array called digest. If fname is NULL, stdin is
   assumed.
*/
void hashfile(char *fname, unsigned char* digest)
{
    int bytesread, i;
    sapphire *hasher;
    unsigned char buffer[16384];
    FILE* f;

    memset(digest, 0, HASHSIZE);
    if (fname)
    {
        f = fopen(fname, "rb");
        if (!f)
        {
            fprintf(stderr, "Can't open %s\n", fname);
            return;
        }
    }
    else
    {
        f = stdin;
    }
    hasher = new sapphire;
    if (!hasher)
    {
        fprintf(stderr, "Out of memory!");
        exit(4);
    }
    while (!feof(f))
    {
        bytesread = fread(buffer, 1, 16384, f);
        for (i = 0; i < bytesread; i++)
        {
            hasher->encrypt(buffer[i]);
        }
    }
    hasher->hash_final(digest, HASHSIZE);
    if (fname)
        fclose(f);
    delete hasher;
}

/* Convert ASCII hexidecimal digit to 4-bit value. */
unsigned char hexval(char c)
{
    unsigned char h;

    c = toupper(c);
    if (c >= 'A')
        h = c - 'A' + 10;
    else
        h = c - '0';
    return h;
}

/* Verify a file created with sapsum by redirecting output
   to a file. */
int verifyfile(char *fname)
{
    int j, k;
    int found = 0;
    unsigned char digest[HASHSIZE];
    unsigned char expected_digest[HASHSIZE];
    FILE *checkfile;
    char checkline[LINESIZE];
    char *s;
    unsigned char err;

    checkfile = fopen(fname, "rt");
    if (!checkfile)
    {
        fprintf(stderr, "Can't open %s\n", fname);
        return(0);
    }
    do
    {
        s = fgets(checkline, LINESIZE, checkfile);
        if (s)
        {
            if ((strlen(checkline)>((2*HASHSIZE)+4))&&
                (!strncmp(checkline,"==>", 3)))
            {
                /* Overwrite newline. */
                checkline[strlen(checkline)-1]=0;
                found = 1;

                /* Read expected check value. */
                for (k=0, j=3; k < HASHSIZE; k++)
                {
                    expected_digest[k]=hexval(checkline[j++]);
                    expected_digest[k]=(expected_digest[k]<<4)
                        +hexval(checkline[j++]);
                }

                /* Compute fingerprints */
                s = checkline+((2*HASHSIZE)+4);
                hashfile(s, digest);

                /* Compare fingerprints */
                err = 0;
                for (k=0; k<HASHSIZE; k++)
                    err |= digest[k]- expected_digest[k];
                if (err)
                {
                    fprintf(stderr, "FAILED: %s\n"
                        " EXPECTED: ", s);
                    for (k=0; k<HASHSIZE; k++)
                        fprintf(stderr, "%02X",
                            expected_digest[k]);
                    fprintf(stderr,"\n    FOUND: ");
                    for (k=0; k<HASHSIZE; k++)
                        fprintf(stderr, "%02X", digest[k]);
                    fprintf(stderr, "\n");
                }
                else
                {
                    printf("OK: %s\n", s);
                }
            }
        }
    } while (s);
    return found;
}



void syntax(char *progname)
{
    printf("\nsyntax:\n"
     "%s [-c|-h][-q] file name[s]\n"
     "    -c = check files against previous check values\n"
     "    -g = generate Sapphire II check values (default)\n"
     "    -h = display this help\n"
     "For example,\n"
     "sapsum test.txt > check.txt\n"
     "generates check value for test.txt in check.txt, and\n"
     "sapsum -c check.txt\n"
     "checks test.txt against the check value in check.txt\n"
     "Written by Michael Paul Johnson & dedicated to the\n"
     "Public Domain August 2000\n",
     progname);
    exit(1);
}


/**********************************************************/

int main(int argc, char** argv)
{
    int i, j, k;
    int check = 0;
    int found = 0;
    unsigned char digest[HASHSIZE];
    unsigned char expected_digest[HASHSIZE];
    FILE *checkfile;
    char checkline[LINESIZE];
    char *s;
    unsigned char err;
#ifdef __BORLANDC__
    struct ffblk f;
    int done;
    char path[MAXPATH];
    char drive[MAXDRIVE];
    char dir[MAXDIR];
    char name[MAXFILE];
    char ext[MAXEXT];
#endif

    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-')
        {
            switch (argv[i][1])
            {
                case 'c':
                case 'C':
                    check = 1;
                    break;
                case 'g':
                case 'G':
                    check = 0;
                    break;
                default:
                    syntax(argv[0]);
            }
        }
    }

    for (i=1; i<argc; i++)
    {
        if (argv[i][0] != '-')
        {
#ifdef __BORLANDC__
            fnsplit(argv[i], drive, dir, name, ext);
            done = findfirst(argv[i], &f, FA_RDONLY |
                FA_HIDDEN|FA_SYSTEM|FA_ARCH);
             while (!done)
            {
                sprintf(path, "%s%s%s", drive, dir, f.ff_name);
                s = path;
#else
                s = argv[i];
#endif

                if (check)
                {   /* Check fingerprint file. */
                    found |= verifyfile(s);
                }
                else
                {   /* Generate fingerprints & write to
                       stdout. */
                    hashfile(s, digest);
                    printf("==>");
                    for (j=0; j<HASHSIZE; j++)
                        printf("%02X", digest[j]);
                    printf(" %s\n", s);
                    found = 1;
                }

#ifdef __BORLANDC__
                done = findnext(&f);
            }
#endif
        }
    }
    if (!found)
    {
        if (check)
        {
            fprintf(stderr,
                "No Sapphire II hash lines found in %s\n",
                argv[i]);
        }
        else
        {
            fprintf(stderr, "No files checked.\n");
            syntax(argv[0]);
        }
    }
    return(0);
}
