/*
 * sha2test.c
 *
 * Version 1.0.0beta1
 *
 * Written by Aaron D. Gifford <me@aarongifford.com>
 *
 * Copyright 2000 Aaron D. Gifford.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR(S) OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>

#include "sha2.h"

typedef struct _test_vector {
	int		data_len;
	unsigned char	*input_data;
	char		*sha256_result;
	char		*sha384_result;
	char		*sha512_result;
} test_vector;

#define NUM_TEST_VECTORS 14

const static test_vector test_vectors[NUM_TEST_VECTORS] = {
  {
    /*
     * TEST VECTOR #1:
     *
     * This test vector is taken from the PDF document that describes
     * the SHA-256/384/512 algorithms.  That document contains sample
     * output for all three versions (SHA-256, SHA-384, and SHA-512).
     */
    3,

    (unsigned char*)
    "abc",

    "ba7816bf8f01cfea414140de5dae2223b00361a396177a9cb410ff61f20015ad",

    "cb00753f45a35e8bb5a03d699ac65007272c32ab0eded1631a8b605a43ff5bed"
    "8086072ba1e7cc2358baeca134c825a7",

    "ddaf35a193617abacc417349ae20413112e6fa4e89a97ea20a9eeee64b55d39a"
    "2192992a274fc1a836ba3c23a3feebbd454d4423643ce80e2a9ac94fa54ca49f"
  },
  {
    /*
     * TEST VECTOR #2:
     *
     * The PDF document only provided sample output for SHA-256 using
     * this test data.  I have provided SHA-384 and SHA-512 sample
     * output from my own implementation which may not be correct.
     */
    56,

    (unsigned char*)
    "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",

    "248d6a61d20638b8e5c026930c3e6039a33ce45964ff2167f6ecedd419db06c1",

    "3391fdddfc8dc7393707a65b1b4709397cf8b1d162af05abfe8f450de5f36bc6"
    "b0455a8520bc4e6f5fe95b1fe3c8452b",

    "204a8fc6dda82f0a0ced7beb8e08a41657c16ef468b228a8279be331a703c335"
    "96fd15c13b1b07f9aa1d3bea57789ca031ad85c7a71dd70354ec631238ca3445"
  },
  {
    /*
     * TEST VECTOR #3:
     *
     * For this test data (from the PDF document), no example output
     * was provided for SHA-256 (SHA-384 and SHA-512 samples were
     * provided), so the sample for SHA-256 comes from the output of
     * my own implementation and so may not be correct.
     */
    112,

    (unsigned char*)
    "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmno"
    "ijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",

    "cf5b16a778af8380036ce59e7b0492370b249b11e8f07a51afac45037afee9d1",

    "09330c33f71147e83d192fc782cd1b4753111b173b3b05d22fa08086e3b0f712"
    "fcc7c71a557e2db966c3e9fa91746039",

    "8e959b75dae313da8cf4f72814fc143f8f7779c6eb9f7fa17299aeadb6889018"
    "501d289e4900f7e4331b99dec4b5433ac7d329eeb6dd26545e96e55b874be909"
  },
  {
    /*
     * TEST VECTOR #4:
     *
     * The output samples for this test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample comes from...well
     * most anyone in the U.S. will know and many outside the U.S. too.
     */
    1515,

    (unsigned char*)
    "Four score and seven years ago our fathers brought forth on this "
    "continent, a new nation, conceived in Liberty, and dedicated to "
    "the proposition that all men are created equal.  Now we are engaged "
    "in a great civil war, testing whether that nation, or any nation so "
    "conceived and so dedicated, can long endure. We are met on a great "
    "battlefield of that war.  We have come to dedicate a portion of that "
    "field, as a final resting place for those who here gave their lives "
    "that that nation might live.  It is altogether fitting and proper "
    "that we should do this.  But, in a larger sense, we can not dedicate--"
    "we can not consecrate--we can not hallow--this ground.  The brave "
    "men, living and dead, who struggled here, have consecrated it, far "
    "above our poor power to add or detract.  The world will little note, "
    "nor long remember what we say here, but it can never forget what they "
    "did here.  It is for us the living, rather, to be dedicated here to "
    "the unfinished work which they who fought here have thus far so nobly "
    "advanced.  It is rather for us to be here dedicated to the great task "
    "remaining before us--that from these honored dead we take increased "
    "devotion to that cause for which they gave the last full measure of "
    "devotion--that we here highly resolve that these dead shall not have "
    "died in vain--that this nation, under God, shall have a new birth of "
    "freedom--and that government of the people, by the people, for the "
    "people, shall not perish from the earth.  -- President Abraham Lincoln, "
    "November 19, 1863",

    "4d25fccf8752ce470a58cd21d90939b7eb25f3fa418dd2da4c38288ea561e600",

    "69cc75b95280bdd9e154e743903e37b1205aa382e92e051b1f48a6db9d0203f8"
    "a17c1762d46887037275606932d3381e",

    "23450737795d2f6a13aa61adcca0df5eef6df8d8db2b42cd2ca8f783734217a7"
    "3e9cabc3c9b8a8602f8aeaeb34562b6b1286846060f9809b90286b3555751f09"
  },
  {
    /*
     * TEST VECTOR #5:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample is EMPTY (no bits).
     */
    0,

    (unsigned char*)
    "",

    "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",

    "38b060a751ac96384cd9327eb1b1e36a21fdb71114be07434c0cc7bf63f6e1da"
    "274edebfe76f65fbd51ad2f14898b95b",

    "cf83e1357eefb8bdf1542850d66d8007d620e4050b5715dc83f4a921d36ce9ce"
    "47d0d13c5d85f2b0ff8318d2877eec2f63b931bd47417a81a538327af927da3e"
  },
  {
    /*
     * TEST VECTOR #6:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample exactly the same
     * length as the SHA-256 block length.
     */
    64,

    (unsigned char*)
    "This is exactly 64 bytes long, not counting the terminating byte",

    "ab64eff7e88e2e46165e29f2bce41826bd4c7b3552f6b382a9e7d3af47c245f8",

    "e28e35e25a1874908bf0958bb088b69f3d742a753c86993e9f4b1c4c21988f95"
    "8bd1fe0315b195aca7b061213ac2a9bd",

    "70aefeaa0e7ac4f8fe17532d7185a289bee3b428d950c14fa8b713ca09814a38"
    "7d245870e007a80ad97c369d193e41701aa07f3221d15f0e65a1ff970cedf030"
  },
  {
    /*
     * TEST VECTOR #7:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample one byte shorter
     * than the SHA-256 block length.
     */
    63,

    (unsigned char*)
    "For this sample, this 63-byte string will be used as input data",

    "f08a78cbbaee082b052ae0708f32fa1e50c5c421aa772ba5dbb406a2ea6be342",

    "37b49ef3d08de53e9bd018b0630067bd43d09c427d06b05812f48531bce7d2a6"
    "98ee2d1ed1ffed46fd4c3b9f38a8a557",

    "b3de4afbc516d2478fe9b518d063bda6c8dd65fc38402dd81d1eb7364e72fb6e"
    "6663cf6d2771c8f5a6da09601712fb3d2a36c6ffea3e28b0818b05b0a8660766"
  },
  {
    /*
     * TEST VECTOR #8:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample exactly the same
     * length as the SHA-384 and SHA-512 block lengths.
     */
    128,

    (unsigned char*)
    "And this textual data, astonishing as it may appear, is exactly "
    "128 bytes in length, as are both SHA-384 and SHA-512 block sizes",

    "0ab803344830f92089494fb635ad00d76164ad6e57012b237722df0d7ad26896",

    "e3e3602f4d90c935321d788f722071a8809f4f09366f2825cd85da97ccd2955e"
    "b6b8245974402aa64789ed45293e94ba",

    "97fb4ec472f3cb698b9c3c12a12768483e5b62bcdad934280750b4fa4701e5e0"
    "550a80bb0828342c19631ba55a55e1cee5de2fda91fc5d40e7bee1d4e6d415b3"
  },
  {
    /*
     * TEST VECTOR #9:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample is one byte shorter
     * in length than the SHA-384 and SHA-512 block lengths.
     */
    127,

    (unsigned char*)
    "By hashing data that is one byte less than a multiple of a hash "
    "block length (like this 127-byte string), bugs may be revealed.",

    "e4326d0459653d7d3514674d713e74dc3df11ed4d30b4013fd327fdb9e394c26",

    "1ca650f38480fa9dfb5729636bec4a935ebc1cd4c0055ee50cad2aa627e06687"
    "1044fd8e6fdb80edf10b85df15ba7aab",

    "d399507bbf5f2d0da51db1ff1fc51c1c9ff1de0937e00d01693b240e84fcc340"
    "0601429f45c297acc6e8fcf1e4e4abe9ff21a54a0d3d88888f298971bd206cd5"
  },
  {
    /*
     * TEST VECTOR #10:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample is exactly 5 times
     * size of the SHA-256 block length.
     */
    320,

    (unsigned char*)
    "\x17\xef\x62\x1e\x89\xcf\x69\xa2\xde\x78\xa4\x06\x1e\x06\x59\xe8"
    "\x0a\xcf\xc6\xb2\xd2\x04\xa9\x14\x49\x32\x53\xe3\x31\xbe\x78\xd2"
    "\xe1\x06\x8a\x3c\x5c\x3c\x58\xe5\xcb\x45\x59\xf1\x94\x51\xf9\xfa"
    "\x00\x89\x62\x90\x27\x83\x21\x53\xdc\x73\x43\x80\xa8\xf5\x46\xf3"
    "\x91\x28\xdc\x01\xfd\x15\x47\x00\xcc\x3e\x8b\x6d\x3f\x7a\xf2\xdf"
    "\xad\x8c\x4e\x1b\x79\x80\x1b\x6c\x5d\x51\x73\xec\x3c\x65\xf3\x04"
    "\xb9\xbd\x53\xa9\x4e\x8e\xcc\xd0\xe8\xe2\xfb\x03\x24\xeb\x9c\x42"
    "\xa4\x1e\xb0\xec\x56\xf7\x3c\xcd\xc5\xbd\xfd\x40\xf9\x05\x41\xda"
    "\xfa\x6c\x49\xd6\xf1\xe8\x7d\xac\xe0\x00\x43\x0c\xa3\xe8\xc9\x71"
    "\x96\xdf\x5c\x9f\x43\x1f\x7b\xe8\x04\x59\x16\x01\xc8\x7d\x14\xd2"
    "\xd2\x61\xbd\x77\xaf\xcd\xe0\x70\x9a\xf9\x7b\x92\x32\xc2\xd2\xd6"
    "\xb6\x11\x30\x41\xfa\xf6\xe7\xfe\xaa\xaa\xde\xd8\x3f\x5c\xc6\x16"
    "\xf3\xd6\x5d\xbe\x7a\x07\x74\xfa\x41\xaf\x0a\x41\x76\x99\x2f\xd4"
    "\xfa\x5d\x2d\x98\xe4\xb9\x9d\xf5\xbc\x68\xe9\xdf\xed\x3b\xb7\x9f"
    "\x52\xab\x5f\xd5\xa8\x22\x05\x42\xa9\x7a\x4a\xf5\xf0\x0a\xa2\x1f"
    "\xd5\x39\xfb\x31\xe1\xa0\x69\xe6\xda\x41\xe4\x26\xba\x26\x55\xe5"
    "\x45\x79\x28\x57\x47\xce\x0c\xc2\x8c\x39\x71\x18\xa2\xe8\x37\x28"
    "\xd4\x14\xa0\x56\x94\x1b\x7d\xd9\x6c\x3b\x56\x85\xd9\xec\x50\x93"
    "\x50\x97\xbe\xe9\x03\x1a\x3c\x1c\xc5\x80\x65\x26\xff\x32\x5a\x69"
    "\x79\xc5\xe7\x9a\x7b\x86\xb3\xb3\xf8\xe2\x9c\x1b\x1b\xf8\xfa\xb1",

    "a7f001d996dd25af402d03b5f61aef950565949c1a6ad5004efa730328d2dbf3",

    "b8261ddcd7df7b3969a516b72550de6fbf0e394a4a7bb2bbc60ec603c2ceff64"
    "3c5bf62bc6dcbfa5beb54b62d750b969",

    "caf970d3638e21053173a638c4b94d6d1ff87bc47b58f8ee928fbe9e245c23ab"
    "f81019e45bf017ecc8610e5e0b95e3b025ccd611a772ca4fb3dfba26f0859725"
  },
  {
    /*
     * TEST VECTOR #11:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample is one byte less
     * than 7 times the size of the SHA-256 block length.
     */
    447,

    (unsigned char*)
    "\x4d\x7f\x47\x62\x4b\xa2\x5c\x83\xd3\x72\x56\x33\x67\xae\x01\x45"
    "\x18\x4c\xb9\x8c\xce\x50\x45\x74\x69\x6e\x87\xb2\xeb\x34\x2a\x10"
    "\x33\xca\x9c\xe5\x68\xb3\x89\xab\x82\xa9\x26\x85\xbb\x58\xe3\x00"
    "\x1f\x06\xf2\x80\x5f\x61\xa2\xfc\x47\x44\xf0\x41\xe5\x73\xc4\x95"
    "\x93\x96\xbb\xa4\x5b\x63\xe9\x95\xe4\x34\x66\x48\x93\xb7\x33\x74"
    "\xf6\x93\x41\x63\x90\x4c\xd9\x0c\x8f\x61\x49\x72\x18\x31\xd9\x5f"
    "\x03\x58\x66\x01\xa6\x4f\xbc\xa6\xdc\xa8\xca\x04\xe3\x9a\xba\x31"
    "\x8d\xeb\x8a\xd9\xe3\x50\xf8\x74\x5f\xdb\x23\xd2\xd0\xc0\x2f\xf9"
    "\x4b\x9d\x8c\xac\x6e\x23\x08\x0a\xbe\x33\xe7\xc1\xa8\x7b\x3d\x5e"
    "\x7b\x3c\xeb\xe0\x85\x48\x88\x19\x15\x23\x5d\x64\x4b\x10\xa9\xfd"
    "\x7e\xa5\xb2\x99\x37\x5e\xc5\xe0\xc6\xd3\xa8\xdc\x8d\x09\x75\x4d"
    "\x2a\x36\x1f\xb3\x2e\x59\x3b\xed\x97\x1b\x26\x32\xbe\xc0\xda\x21"
    "\x5a\x29\xc7\x66\x92\x78\x84\xc9\x1b\x74\xd3\xc4\x72\x00\x97\x43"
    "\x80\x15\x19\x6f\x89\x1a\x0e\x01\xf9\x87\xc2\x19\x7a\xce\xd4\xad"
    "\x89\x2a\xc2\xd2\x1d\x64\x69\x70\xd6\xa0\xbc\xb3\xd8\x80\x72\xf9"
    "\xaf\xf5\x94\xe8\xa5\x25\x95\xea\x07\x26\x6c\xb8\xc3\x3d\x27\xa0"
    "\xa0\xf1\x20\xae\x94\x09\x56\x6a\xce\x02\x9e\x31\x93\xee\x1e\x42"
    "\x07\x46\x96\x29\x27\x76\x4e\x00\x24\x51\xd9\x30\xf1\xc2\x9e\x5a"
    "\xf0\x6a\x7e\xe8\x4c\x44\x2e\x48\x3b\x46\xdc\xab\xb2\x93\x95\x46"
    "\xce\xa9\xf7\xc7\x5d\x66\xc7\x86\x55\x0f\x90\xba\x7b\x7b\x40\x0c"
    "\xde\xf9\xb0\x62\xc6\x29\x75\x51\xee\x47\xd5\x58\x6b\x8e\xd6\xcd"
    "\xad\x02\x5b\xcc\xfd\x62\x0b\xd2\x11\x25\x5b\x5d\xe2\xd0\xc7\x3e"
    "\xec\xae\x97\x89\x6a\x7e\x99\xfd\xef\xa3\xb9\x46\x12\xc9\x6c\x3b"
    "\xef\x33\x85\x68\x72\x18\xcb\x72\x26\x48\x6b\x29\xb6\xfc\x2d\xf3"
    "\xe6\xdd\xa0\x53\xb7\xf9\x0d\xea\xab\x91\xcf\x5f\x19\x16\x8b\xe1"
    "\xa2\x0f\x74\x90\x62\xef\xee\x62\x3d\x40\x18\xeb\xe2\x7a\x88\x63"
    "\x2a\x69\xed\xda\x78\x81\xe7\x4e\xf8\xcb\x5e\xe8\x71\xa7\xd6\x6a"
    "\x89\x35\xb7\xd4\x3a\x28\xe5\x1e\x12\x27\x2c\x35\xd5\x94\xb4",

    "6dcd63a07b0922cc3a9b3315b158478681cc32543b0a4180abe58a73c5e14cc2",

    "548e4e9a1ff57f469ed47b023bf5279dfb4d4ca08c65051e3a5c41fab84479a2"
    "05496276906008b4b3c5b0970b2f5446",

    "ee5d07460183b130687c977e9f8d43110989b0864b18fe6ee00a53dec5eda111"
    "f3aaa3bac7ab8dae26ed545a4de33ed45190f18fa0c327c44642ab9424265330"
  },
  {
    /*
     * TEST VECTOR #12:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample is exactly 5 times
     * size of the SHA-384 and SHA-512 block lengths.
     */
    640,

    (unsigned char*)
    "\x75\x90\xa2\x7b\x2a\x74\x28\x10\x89\xaa\x31\x82\x90\xb7\x29\x64"
    "\xa7\x0b\x09\x41\x08\x98\x18\x8b\x23\xe1\x03\xb8\x64\xe4\xa9\xe4"
    "\x97\x08\xc9\xdc\x3d\x4c\x32\x69\xe6\x3e\xd3\x66\xed\x00\x1d\xf0"
    "\x44\xb4\x01\x95\x33\x72\xf1\xce\x97\x4e\xcf\xec\x03\x56\xa6\xa4"
    "\x26\xf9\xdf\x6b\x79\x17\x4a\xfe\x65\x27\xb5\x73\x09\xa1\x5b\x29"
    "\xfb\x96\xed\x16\x2c\x70\xbc\x8b\x9b\xdd\x9d\x29\x12\x55\xdb\xf2"
    "\xe8\x02\x9e\x51\xaa\xaa\xc5\xdb\x21\x9f\xf0\xff\x1f\x51\xb8\xa6"
    "\x44\x88\xcc\x2c\x7e\x6b\x3b\xf7\x88\x39\x29\x44\x2f\xb0\x73\x2d"
    "\xaf\xce\x43\xc9\x3b\x42\xdb\xbf\x74\x99\xdc\x61\xab\xfd\x80\xb1"
    "\xf5\xaa\x78\x15\x38\x7c\x6f\x8b\xad\x54\x55\x4b\x56\x32\xa7\x75"
    "\xf9\xc9\x5c\xed\xcf\x44\x00\xb3\xf7\x83\x95\xd4\xe2\xf5\x9c\x0f"
    "\x6f\x05\xdb\x72\xc0\x41\x8a\x94\x10\x59\xb6\x5f\xf0\x3b\x8e\x31"
    "\x86\x0b\xbd\xe0\xc5\x8b\x7f\xec\x0c\x1f\xae\x87\x3b\x32\x2a\xe9"
    "\x46\x6a\x3f\xf4\x39\x11\xfc\xcf\xe9\xea\x70\x87\x19\x2a\xee\x07"
    "\x53\xbd\x23\xdf\x63\xdd\x83\xb0\x36\xe4\x71\x4d\x8e\x5f\xe5\x91"
    "\x76\x5f\x0f\xb6\x1f\x2d\xba\x40\xd1\xab\x75\x59\x47\xd8\x1e\x45"
    "\xa3\xee\x00\xdf\xbe\x64\x85\x23\xf6\x26\x4a\xa1\xe7\xf9\xe0\x6e"
    "\x67\x88\x34\x93\x1c\xdd\x5a\x21\x1d\xf3\x4c\xf5\x14\x7d\xf1\xf8"
    "\xb8\x5e\xda\xd9\x55\x73\xd1\xb1\xd2\x0f\xdc\x5b\xf3\xd2\x71\x6a"
    "\xe7\xb2\x73\xdb\x4a\x81\x99\x92\xb9\x8b\x4e\x04\x9b\x32\xb8\xe8"
    "\xf6\xc6\xc0\x73\xf0\x35\x29\x42\xaf\x1e\x52\x88\xe9\x9a\x17\x2a"
    "\x32\x17\x74\x17\x03\x2b\x53\xfc\x55\xed\x60\x6d\xba\xb2\xac\xe2"
    "\x0e\xbd\x80\xa2\x1b\x08\x76\xd5\xb7\xaa\xb5\x38\xf4\x27\x81\xdc"
    "\x3d\xef\xea\x7b\xbd\x1a\x7f\xe0\xed\x5f\xb7\xc4\xef\xc3\x47\xd9"
    "\xdb\x42\xb9\x92\xf6\x03\xfc\x30\xc3\x7c\x09\xa9\x0e\xe4\x9d\x9d"
    "\x80\x53\x37\x9e\xc6\xf5\xe8\xa6\x97\xdf\x8d\x40\x64\x0b\x5c\x6c"
    "\x25\xcc\x41\x83\x89\x98\x0d\x34\x9b\x2f\x51\x7c\xa2\x57\x60\x77"
    "\x5c\xdc\xc0\x17\x97\x29\xe4\x96\x17\x30\x27\x5e\x90\xd0\x02\x69"
    "\xe9\xab\xf8\xa5\xd4\x1d\xdb\x56\x40\x96\xa5\x51\x2f\x92\x17\x79"
    "\x44\x2a\x8a\xe4\xc1\x05\xa0\x80\xc5\x2f\xce\x86\x6b\xd1\x71\xb0"
    "\xca\x3c\x04\xcc\xf8\x24\x22\x63\xc3\xc2\x1a\xdb\x51\x68\xb8\x1a"
    "\x6b\xa0\x18\x38\xce\x8f\x6d\x3a\x5b\xe4\x7d\xa4\xcc\xba\x71\x02"
    "\x33\xe7\x02\x26\xff\xed\x6d\x23\x9b\x31\xbd\x8e\xc4\x8c\x84\x18"
    "\x28\xed\x84\x26\xe5\x1e\x66\x15\x3c\xd6\x93\x5b\x45\x93\xc2\xde"
    "\x54\x79\x82\x0a\x74\x0c\x25\x16\x51\x00\xa7\xc3\x91\x3b\xd4\xa9"
    "\xf3\xe5\x48\x7d\x60\x74\x79\xc2\xfc\x95\xf7\x37\x84\x48\x1c\x11"
    "\xab\x50\x80\x36\x9a\x96\x8a\x36\x38\xa5\xa5\xe0\xdf\x99\x32\xa3"
    "\x65\x67\x66\xbe\xc9\x04\x66\x7f\x72\x43\x8f\xd4\x9c\xd5\x15\xb0"
    "\x9e\x35\xb8\x1d\x65\xc3\x78\xb2\x38\x80\x87\x27\x8f\x86\xc7\x84"
    "\xc0\xc0\xa7\xd3\xe6\x30\x61\x7d\xdb\x9c\xe3\xa9\x3e\x16\x09\x49",

    "af6ebfde7d93d5badb6cde6287ecc2061c1cafc5b1c1217cd984fbcdb9c61aaa",

    "c6fec3a3278dd6b5afc8c0971d32d38faf5802f1a21527c32563b32a1ac34065"
    "6b433b44fe2648aa2232206f4301193a",

    "73ffeb67716c3495fbc33f2d62fe08e2616706a5599881c7e67e9ef2b68f4988"
    "ea8b3b604ba87e50b07962692705c420fa31a00be41d6aaa9f3b11eafe9cf49b"
  },
  {
    /*
     * TEST VECTOR #13:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample is one byte short
     * of 17 times size of the SHA-384 and SHA-512 block lengths.
     */
    2175,

    (unsigned char*)
    "\x6f\x9b\x06\x80\x71\x1a\xd2\xb4\x88\x75\x72\x4f\x05\xb4\xca\xd5"
    "\xda\xe8\x08\x81\x6e\x04\xa5\xb5\x99\x08\xc2\xec\x0e\x91\xe6\xf2"
    "\x32\xd5\xd8\xd7\xea\x9a\xca\x0c\xf4\x62\x4d\x2e\x6e\x9c\xf4\xf7"
    "\x50\x36\x45\x92\x75\xa7\xb4\xb6\xc8\x6e\x2a\xaf\x46\xad\xf3\xb0"
    "\xb0\x3a\x59\x73\xeb\xe7\xc8\x55\x13\xa9\xd2\xcf\x55\x39\xe1\xd9"
    "\xca\xb1\xd2\xab\xd2\x95\xec\xe4\xe0\xfe\x47\xdd\xb1\xdd\x1a\x71"
    "\x6b\x68\xad\x38\x3a\x84\x20\x3b\x9c\xb6\x58\xae\x02\x9d\x0d\x5f"
    "\x17\x08\x9d\xaa\x13\x1b\xdc\xa4\xf6\xf2\xa9\x61\x70\xe7\x70\x60"
    "\x1c\x71\x9a\x25\x78\xd6\xb7\x00\x40\x98\xf5\x20\x81\xf2\x95\x26"
    "\x81\xda\xc3\x3a\x19\x2c\xb3\xe1\xb8\x89\x16\xc6\xaa\xee\xd0\xd0"
    "\x76\x47\xe5\x4e\x31\xb5\xd4\xe1\x1b\x62\x73\x1c\x32\x4e\x68\x1b"
    "\x27\xdd\xea\x84\xaa\xe8\xa4\x6a\x44\xad\x97\xc5\x1d\x7f\x99\x0c"
    "\x39\x15\xa0\x18\xeb\xf5\x10\x50\x05\xeb\x73\x99\x59\xb3\xb4\x36"
    "\x73\x1e\xef\xf9\x8c\xdd\xfc\x2a\xb0\x16\xe8\xc9\x63\x62\x03\xbc"
    "\x7a\x5b\x1b\x45\x1d\x39\xcb\x7a\x09\x63\xb8\x42\x7d\xaa\xed\x8a"
    "\xee\x17\x21\xd5\xde\x79\xce\xfc\x3d\x9d\x59\xfc\xd6\x56\x5c\xa0"
    "\xb6\xa6\x4c\x01\x97\x68\x94\x47\xef\x18\x43\x13\x75\x44\xc6\x00"
    "\xfd\xdc\x04\x7d\xe1\xcf\xc4\x19\x49\x1e\x11\xca\xe7\x21\xf7\x33"
    "\x51\x52\x53\x58\x61\x6e\x1e\x68\x10\x3f\x2a\x1d\x65\xdc\xca\x55"
    "\xe1\xa5\xb5\x1e\x2c\xf7\x38\x0f\x62\x09\xbe\x62\x53\x6f\x83\x89"
    "\x46\xdf\x50\x4f\xed\x4e\x1c\x3c\x48\x6b\xc6\x87\xa4\xa5\x49\x62"
    "\x85\xec\x3f\xa2\x19\x54\x82\x84\x10\x92\x0e\xe8\x41\x35\x6f\x8f"
    "\x1d\x92\x65\xb4\x53\xfd\x60\x55\x5c\x1a\xbf\x18\xa0\xd5\xc3\x33"
    "\x56\x5f\x92\x8a\xed\x4f\x1e\x91\x11\x40\xc3\x2e\x9a\x3a\xdb\x22"
    "\x38\x6f\x85\x26\x43\x82\x55\x12\xe5\xe1\x1e\x7c\xe7\x98\x37\x08"
    "\x68\x22\x62\x49\x88\xf9\xd5\xb4\x69\x95\x42\x41\x0f\x1c\x68\x81"
    "\x49\x93\x6a\x90\xc4\x5f\xe6\xa9\x57\x9f\x30\x19\x7b\xc0\x64\xa4"
    "\xb8\x99\x76\xdc\x78\xb1\x83\x58\xcc\x7c\xd6\x8e\x76\x7a\x3b\x83"
    "\x09\xc7\xb2\x08\xf4\x94\x55\x2c\xc6\xea\x70\xbb\xb0\x0b\xc3\x8e"
    "\x8e\x60\xa5\x7d\xe2\xfd\x6f\x11\xce\x65\xa2\xdd\x3d\xde\x61\x80"
    "\x23\xde\x20\x5c\xb6\x88\x8e\xad\xac\xe1\x80\x90\x84\xdc\xcd\x5f"
    "\x5b\x9e\x2f\x93\xc9\x94\x30\x81\x19\xbb\x51\xc2\x16\x41\xcb\x14"
    "\x46\xf8\xed\x8d\x1a\x56\xee\x85\x78\x05\x7c\x1d\x83\xa6\x18\x3d"
    "\xfb\xa4\xca\x0c\x9e\x07\xae\x46\x5e\xd6\x06\x83\xac\x8b\x55\x1d"
    "\xb7\x71\x97\xb1\x27\x02\x7d\x23\x72\xfe\x65\x31\xc6\xf9\x14\x66"
    "\x23\x8d\xe4\x98\xac\xb8\x3b\x98\x30\xd1\xbc\x00\x44\x95\xd1\xc7"
    "\x54\xc9\x99\x7a\x85\x4a\xee\x55\xf0\x57\xe9\xbb\xc8\x3d\xa3\xed"
    "\x6a\x26\x9f\x75\x59\xcc\x25\x74\x55\x63\xfe\x0b\x2f\x1d\xd2\x3e"
    "\xe3\x81\x7c\x7d\x3e\x30\x06\xcb\x16\x92\xa9\x47\x61\x6e\x4a\xb2"
    "\xe0\x6f\xc6\x91\xe5\x13\xb7\xf9\x9e\x9c\xbb\x8b\x5f\x38\xf5\x57"
    "\xd7\xa9\x27\xdd\x34\xfe\x2f\x3c\x0e\x52\xa2\xe4\xb7\x4e\xb9\x1b"
    "\xfa\x12\x8a\x29\xd1\x8b\x09\x7e\x1a\xb3\x7c\x2d\x23\xe2\xfa\x08"
    "\xbe\xdb\xc0\x34\x0d\x99\xe8\x0d\x3c\xad\x96\x62\xa0\xa0\xc6\x74"
    "\x41\x39\x37\xef\x4f\xd1\x55\xd9\x71\x9a\x18\x6e\x21\xed\xbf\x48"
    "\x6d\x33\x8d\x28\xb2\x26\x3f\x59\x48\xd1\xde\x8a\xd4\x2d\xdf\xa4"
    "\x15\x7e\x15\x36\x50\xfb\xe5\x96\xc1\xd0\x12\xa9\x63\xe9\xfe\x9e"
    "\x3b\xb6\x49\x47\xc6\x7b\x8e\xda\xc1\xa5\xa0\x78\x86\x30\xca\xd4"
    "\xad\x59\xef\x63\x46\xd8\x21\x9c\x5d\xc6\xdc\x91\x06\x3f\xe9\x2e"
    "\xb1\xd6\x97\x91\xbb\xe1\xf8\x4d\xe4\x4f\x53\x5d\x73\x31\xd4\x1e"
    "\x8e\xf6\x17\x7c\xef\xc8\x7b\x5b\xf3\x20\x2f\x16\x84\xcd\x79\x7b"
    "\x76\x4e\x21\x0f\x71\x96\xb4\xe1\xe8\x89\x4e\x80\x46\x4f\x72\x26"
    "\x80\x56\x83\x9f\x4f\x61\x19\x0d\x9d\x64\xe7\x1b\x2d\xc4\x26\xdb"
    "\x94\xd1\xb9\x64\x15\xc2\x02\x2c\xa9\xc8\xd0\xfa\xb1\xb6\x4d\x6f"
    "\x72\x88\x98\xe4\xbf\x34\x35\x2b\x8e\x7c\x11\x9d\xe8\x57\x5a\xbb"
    "\x06\xcf\xdb\xc5\xa5\x97\x99\x76\x7e\x91\x5e\xd5\x89\xe4\x81\x6e"
    "\x84\x70\x81\x86\x3a\x41\x96\x02\xef\xec\x5b\x96\xa4\x45\x60\xad"
    "\x1d\xaa\x54\x40\x89\xca\xe2\x77\xc1\xe4\x05\xae\x50\x16\x03\xa8"
    "\x0e\xbe\x50\x69\x0e\xd5\x00\xb3\xcc\x2e\x89\x42\x07\xc7\x3a\xec"
    "\xf2\x06\x0f\x12\xf1\x93\x7d\x33\x0d\x71\x3d\x48\xac\xc3\x1e\xdf"
    "\xf6\x64\x18\xd3\x80\x0a\x2a\x27\xd6\x09\x65\x13\xa0\x3a\x1b\x41"
    "\x89\x52\x7c\x87\x05\xdc\x4e\x7b\x0c\xab\xf1\x8a\xde\xe4\xb2\xf9"
    "\xc6\x72\x34\x15\x75\xc6\x6b\x9e\x36\xa3\x6a\xab\xe3\x8e\x1f\x71"
    "\x0e\x5f\xc2\x22\xbb\x9a\xfd\xc4\x49\xf3\xb5\x83\x26\x22\xd3\xeb"
    "\x2b\x40\x42\xf5\x0e\x59\x03\x66\x86\x91\x6d\xaa\x44\x70\xa0\x09"
    "\x2f\x27\x42\x3b\xf8\x81\x01\x28\xe4\x12\x10\xa8\x03\x4c\x4b\x08"
    "\xe4\x03\xba\x47\x1b\x31\x14\x91\x68\x77\xa6\x92\xd7\x20\x8f\x04"
    "\xaa\x69\x4d\xe7\x21\xc2\xf5\x90\x87\xa3\xaa\x6c\x9b\x82\x19\x7e"
    "\xc7\x2c\x77\x79\xd1\xa0\xc8\xd4\x2d\x9d\x8d\xa3\x67\x5b\x8a\xdf"
    "\xfa\xe5\xa5\xcc\x5e\xc7\x4e\xec\x42\xb9\xa1\x14\x38\x0a\x28\x8c"
    "\x0a\xc9\x97\xfc\xe0\x19\x06\x74\xe8\xb4\x6e\xae\x2f\xbd\x16\xb1"
    "\x4c\x97\xfd\xc8\xa3\x25\x1c\x0e\x87\xd9\x05\x41\x7b\xee\xfe\xff"
    "\x5d\x2f\xa1\xd0\x27\xc2\xf0\x21\x70\xf1\x8f\xe1\x4d\x14\xdc\xd7"
    "\x45\xef\x96\x3d\x88\xda\x03\x55\x86\xee\x4d\x72\x8a\x80\x0f\xad"
    "\x2d\x3f\x9c\x2e\xab\x53\xa4\x55\x47\x7c\x92\x1b\x22\x93\x9a\x4c"
    "\xf0\x17\xf6\xe7\xfc\x2a\x2e\x54\xfc\x12\xd4\x53\x44\x90\x0c\xe0"
    "\x32\x76\x7a\x4b\x08\xdf\x63\x52\x08\xb0\x75\xdb\xd0\xce\xb1\x87"
    "\x88\x83\x33\xd6\x40\xf6\x8e\x4b\xb9\xe7\x21\xc4\xd9\xf2\xe4\x61"
    "\x8f\x2e\x55\x41\xef\xc1\xad\xe1\x45\x5f\x44\x4f\x88\x83\x2d\xb4"
    "\x00\x45\xd3\x71\xc4\x8a\xfb\x30\x6a\xfa\xe7\xe8\xc8\xbe\xe4\x91"
    "\x3a\x46\x6c\x8c\xe2\x87\xfe\xaf\x9b\xa9\xe7\x4a\x78\xf4\x0d\x52"
    "\xca\x86\x46\x39\x7b\xf2\xdf\x41\xb4\x08\x38\x5c\xb5\x38\x88\xca"
    "\x43\xc1\x4c\xe5\x6a\x75\x77\x98\x0a\x38\xc9\x53\x16\x58\xe1\x4c"
    "\xd5\x93\x59\x92\xcc\xfc\x59\x4d\x81\x77\xbd\x02\xe6\x80\x7a\x04"
    "\x04\x57\x1a\x9c\xb9\xe9\x34\x20\xbd\x76\xb7\x95\xb3\xf8\x2e\x80"
    "\x1f\x37\x5a\x75\xd4\x35\x22\xc2\xb8\x50\x96\x26\xf4\xf5\xec\xe7"
    "\x9f\x12\x8a\x93\x81\xc1\x12\x17\xb2\xde\xde\xb0\x1f\x12\xc2\x77"
    "\xd6\xdc\x1a\x3c\xe8\x03\xe3\xef\x6a\x57\x7d\x33\x80\xd8\xf0\x9e"
    "\xa2\x27\xa4\xf2\xee\xa9\x8a\x4e\xbb\x88\xee\x8d\x1b\x02\x5a\xbe"
    "\x08\xea\xf1\x74\x2d\x6b\x07\xd8\x85\xbc\x0b\x31\x07\x8f\x89\xb5"
    "\x37\xbf\x35\x3d\x98\xd8\x24\x81\xfe\x69\xb4\x59\xad\xeb\x40\x69"
    "\xaf\x31\x0b\xcd\xa7\xce\xc8\xa3\x56\xd4\x8d\xce\xf7\x54\xe9\xdc"
    "\xf0\xef\x04\x42\x1d\xd6\x5f\x59\x0c\xd7\x92\x3c\x8a\x6d\x4d\xff"
    "\x63\x00\x4e\x10\x13\x04\x94\x7f\x1d\x85\x85\x12\x8b\xd5\x34\x66"
    "\xb2\x89\x55\x64\xff\x40\xbc\xf7\xf2\x17\x19\xaf\x13\x9a\x76\xeb"
    "\xea\xad\xb5\x0c\xdb\xe8\x30\x1d\xd3\xb7\x7c\xcd\xf8\xfe\x2a\xf6"
    "\x86\x84\x45\x2b\xe8\x3b\xab\x75\x21\xad\xc8\xb3\x18\xbf\x49\x58"
    "\xd2\x4b\x8c\x46\x64\xf7\x70\x00\x90\xcc\x4b\x13\x24\xb3\x8a\xb8"
    "\x18\x86\xfc\xf6\xcc\x9f\x48\x0e\x4b\xcb\xcc\x80\x56\x25\x72\x0e"
    "\xb4\x0e\x16\x20\xd0\x94\x86\x24\x9e\x6f\xb1\x3f\x48\x89\x9b\x78"
    "\x75\xd1\x41\x4b\x47\x3c\xbd\x48\x3d\x61\xf6\x94\xd7\x67\x1f\x33"
    "\xd2\x72\x3d\x17\x01\x13\x4c\x24\x8e\xf1\x08\x68\xa1\x25\xd7\x61"
    "\xd3\xa8\x1f\xda\x05\x2a\x6a\x43\xde\xc2\x50\x18\x8b\x31\x5d\x5c"
    "\x7b\xb5\xc1\xb1\x53\x71\x22\xc2\x54\x22\xbf\x9c\x4c\x87\x8c\xf1"
    "\x87\x7c\x59\x08\x38\x4a\xc0\xbc\x2a\xfc\xc5\xa8\xec\x5d\x87\x96"
    "\x9e\x64\x73\x76\x87\xd9\xf0\x99\xf7\x30\x62\xf5\xe9\x56\x4f\x92"
    "\x3f\x8c\x33\xf4\xa2\x22\xae\x55\x09\xd5\xd9\x18\x77\xb8\xc5\x28"
    "\x4c\xaa\x31\x01\xe5\x8a\xbd\xe7\xc4\xe0\xe9\x1d\x9b\xc0\xd7\xde"
    "\x37\xb1\x86\xbc\xe5\x97\x60\xb8\x39\xcf\x27\xf4\xdb\xa5\x64\x90"
    "\x69\x6d\x8b\xb5\xe0\xed\x39\x1d\x2a\x80\x21\xaf\xb2\xb9\xa1\x12"
    "\x55\x77\xe1\xf7\x2a\xc7\xec\xbc\x0a\xbc\x27\x3f\x03\x4d\x16\x36"
    "\xf4\x04\x58\x1f\x64\x0c\x7d\xbd\x9d\x64\x31\xad\x6c\xed\x4a\x61"
    "\x37\x51\x94\xce\x98\x6b\x5d\xb1\x82\x72\xd6\x56\x3c\xeb\x7a\x4a"
    "\x40\x33\x14\x09\x3b\x7a\x71\x41\xf3\xeb\xd2\xf1\xee\xc0\x99\x91"
    "\x5c\x96\x81\x71\xcf\x64\x7e\x92\xad\xda\x45\xfc\x80\x35\x1d\xfb"
    "\x91\x59\xb8\xd8\x9d\xa2\x1e\x84\x8a\xab\x74\x69\x7a\x82\xed\x42"
    "\x25\x7c\xa0\x8e\xdd\x9f\xd4\x82\x20\x52\x94\xb1\xa3\xe2\x33\xdd"
    "\x8f\x14\xd1\x61\xa3\x3b\x0a\x93\xaf\xbf\x08\x50\x04\x22\x82\x21"
    "\x85\x6e\xf8\x18\x8f\x88\xcf\x4d\xfe\xfc\x0f\x60\x23\x0c\x9e\xa6"
    "\x43\x56\x77\x9f\x69\x21\xb3\x93\x81\xfa\x1b\xff\x93\xea\x3a\x1b"
    "\xe8\xaa\x95\xde\xec\x5b\xbb\x84\xbc\x64\x1c\x4d\xaa\x26\xca\x62"
    "\x98\xb2\xdb\xf2\x04\xc4\x7c\xd7\xab\xd2\xef\x48\x02\x6c\xc0\x4a"
    "\x48\x2a\xeb\x49\x25\x95\x86\x0c\x19\xc9\xb2\xe7\xc5\xfb\xc5\x38"
    "\x6b\xee\x8e\xd1\x06\x96\xc3\xcc\xb8\x5e\x08\x91\xf0\xae\x5c\x9d"
    "\x64\x22\x94\x7f\x4a\x71\xb1\x4f\xbe\x12\x83\xc2\x71\x1e\xeb\xd3"
    "\xad\xda\xfd\x66\x87\xa4\xd4\x0e\xe0\xc1\x5c\xba\x6c\x63\xdd\x9d"
    "\x12\xe9\x6f\x04\xa2\x08\x1c\xa6\x83\x9c\xf0\x21\x9e\x1a\xea\xc9"
    "\x11\xad\x3b\x32\x54\xfa\xa6\x20\x20\x6f\x88\xa6\x31\xf7\x7a\xd6"
    "\xc0\x29\x18\x63\xfc\x6d\x08\x47\xa0\xd0\xd8\x07\xbe\xc7\xe6\xcc"
    "\x7f\x95\x6e\x9c\x06\x28\xfd\xea\x40\x3a\xa1\xe1\x4a\x06\xd7\x7e"
    "\x2a\x1a\xd0\xe5\x8b\x07\x47\x56\x07\x62\x79\x9b\x80\xcc\x4e\xde"
    "\xb2\x9d\x5a\x8d\xb4\x3a\x68\x22\x56\x92\x68\x03\x85\x96\x35\x83"
    "\x3d\xab\x4c\x04\x86\x28\x0d\x53\x53\x99\xed\x6d\x91\xe2\x97\xd6"
    "\x06\xbb\x42\x25\xa0\x93\x9f\x6a\x0c\xc4\x43\x4c\x9d\xcc\xc1\x92"
    "\xa7\x76\xf8\xe7\x01\x50\x2b\x3a\xc2\x33\xcd\xe1\x0f\x23\xd0\x30"
    "\x64\x2a\x5c\xf8\x86\x8e\x35\x0c\x0a\x12\xaf\xff\x3c\xa9\x07\xba"
    "\x7e\xac\xee\x8a\xe2\xad\x9d\xf9\xbe\x54\x23\xf9\x6f\x5a\xd9",

    "8ff59c6d33c5a991088bc44dd38f037eb5ad5630c91071a221ad6943e872ac29",

    "92dca5655229b3c34796a227ff1809e273499adc2830149481224e0f54ff4483"
    "bd49834d4865e508ef53d4cd22b703ce",

    "0e928db6207282bfb498ee871202f2337f4074f3a1f5055a24f08e912ac118f8"
    "101832cdb9c2f702976e629183db9bacfdd7b086c800687c3599f15de7f7b9dd"
  },
  {
    /*
     * TEST VECTOR #14:
     *
     * The output samples for thi test vector come exclusively from my
     * own implementation and so may be completely incorrect.  Use with
     * a very large grain of salt.  The input sample 4 KB of misc.
     * data.
     */
    16384,

    (unsigned char*)
    "\xd1\xbc\x8d\x3b\xa4\xaf\xc7\xe1\x09\x61\x2c\xb7\x3a\xcb\xdd\xda"
    "\xc0\x52\xc9\x30\x25\xaa\x1f\x82\x94\x2e\xda\xbb\x7d\xeb\x82\xa1"
    "\x0f\x10\xf2\x77\x27\xa9\x83\xf8\x59\x15\x4e\x68\x73\xc2\x8f\x1c"
    "\x3f\xba\xae\xfd\x26\x40\x19\x14\x8b\x2c\x73\x25\x9b\x41\xfc\x79"
    "\xf0\xf2\xdb\x55\xc7\x54\xe5\xcb\xa2\xf8\x96\x37\x4d\x7f\x84\x6a"
    "\x8d\x3f\xff\xdd\xf7\x9e\x9a\x23\x2f\xfd\x19\xf9\xcc\xaa\x4d\x6b"
    "\x37\xa6\xa2\x43\xdb\xe0\xf2\x31\x37\xb1\x08\xa0\x43\xd9\xda\x13"
    "\x12\x1a\x9b\x50\x5c\x80\x49\x56\xb2\x2e\x93\xc7\xf9\x39\x69\xf4"
    "\xa7\xba\x8d\xde\xa4\x5b\xf4\xaa\xb0\xbe\xbc\x8f\x81\x4e\x09\x91"
    "\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52\x4d\x20\x25"
    "\x36\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52\x4d\x20"
    "\x25\x36\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52\x4d"
    "\x20\x25\x36\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52"
    "\x4d\x20\x51\x32\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f"
    "\x52\x4d\x20\x51\x32\xf0\x94\x66\x38\xf0\xc6\x7f\xe3\xc1\xfb\x6f"
    "\x85\x39\x89\x11\xfc\x06\xf6\xb2\xb1\xae\x58\x56\x35\x53\x33\x0b"
    "\x46\x77\xbc\x0d\xf5\xe3\x9e\x1c\x9a\xdd\x06\x28\xe1\xee\xf7\xba"
    "\xaf\x13\x3a\x5e\xa6\x48\x7a\xf0\x16\xf2\x6d\xf7\x00\xb9\xed\xeb"
    "\xe1\xb1\x48\x12\x78\x4e\x36\x9b\x11\x55\xb4\xf6\x6d\xd2\x83\x33"
    "\x42\x59\xe5\x86\xab\xe7\x6e\xdb\x40\x19\xb1\x31\x2b\xb6\x50\x3f"
    "\x98\xd0\x69\xf0\xfd\xc9\xd4\x55\x52\xe0\x91\x1d\x9e\x50\x49\x72"
    "\x31\xaf\x5d\x2e\xd7\x99\x3a\x62\x29\x32\x2c\x11\x26\x38\x66\x1f"
    "\x67\x59\xf3\xe6\xc8\x50\xc5\x37\x37\x3a\xae\x12\x9d\x00\x84\x82"
    "\x67\x38\x6d\x79\xb0\x10\x0d\xd4\x04\xd4\xb0\xbe\xb1\x95\xe9\x1c"
    "\xfd\xc0\xcf\xd8\x95\x87\x34\x42\xfc\x9d\x70\x9c\x31\xdb\x14\xb8"
    "\x4a\xb0\x12\x81\xbc\xfc\xd2\xce\xf9\xf2\x42\xa2\x67\x1e\xe4\xb0"
    "\x45\xc7\xb9\x24\x4a\xb5\x9a\x41\xff\x66\x7b\x7d\x33\xe3\x28\xde"
    "\xc5\xbc\x6b\xbf\xd3\x10\x3b\x86\x4b\x25\xa2\x1b\x39\x54\x7a\x3d"
    "\xee\x71\x69\xae\xe8\xa6\x80\xc2\x44\xed\x31\x33\xb6\x92\x55\x48"
    "\x88\xd4\x6e\xa2\x66\x40\x9d\x68\x81\x84\x15\xd5\x1f\x30\xa9\xbe"
    "\xe3\xb0\xc4\x42\x98\xfc\x1c\x14\x9a\xfb\xf4\xc8\x99\x6f\xb9\x24"
    "\x27\xae\x41\xe4\x64\x9b\x93\x4c\xa4\x95\x99\x1b\x78\x52\xb8\x55"
    "\xcb\xbb\x9d\x5d\xc1\x05\x9e\xd8\x62\x9a\x29\x2a\x36\x7c\xd5\x07"
    "\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52\x4d\x20\x25"
    "\x36\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52\x4d\x20"
    "\x25\x36\x43\x4f\x50\x59\x20\x54\x52\x41\x4e\x53\x46\x4f\x52\x4d"
    "\xda\xb9\x1d\x8e\x0c\x7c\xfd\x2b\xef\x70\x31\xf0\xa7\x98\xbe\x0b"
    "\x56\x1b\xcf\x37\x60\x31\x83\x3d\xc2\x44\xb1\xff\xdc\xb5\x6e\x18"
    "\x7e\x91\x3e\x6a\x31\xc7\xd9\xdf\xc1\xb4\x6c\xb7\x05\x7d\xce\x33"
    "\x51\xf2\x17\x61\xea\xd3\x5a\xf8\x44\xd6\x45\x16\x07\xcb\x4d\x09"
    "\x30\x06\x1e\xfc\x81\x04\x11\xf2\xb8\x85\x25\x82\x98\xb7\x1b\x83"
    "\x23\x2b\x2d\xf5\xda\x21\x0a\x5f\x1d\xc4\x1d\x9d\xba\x70\x88\x29"
    "\x16\xb5\xe4\x74\xf8\xac\x57\x6c\xd2\x47\x3e\x8c\xaf\x81\x09\x24"
    "\xa7\x1f\x90\x53\xbe\x76\x00\xd2\x02\xd8\x5b\x3e\xfc\xb9\xd9\x7c"
    "\x96\xfc\x08\x2d\x90\xcb\xa9\x30\x33\xef\xd9\x4e\x55\x9c\x81\x61"
    "\x68\x89\x03\xe9\xf9\x12\xf2\x78\xbd\xd0\xd7\xb1\x9b\xf9\xc7\x96"
    "\xbb\xe2\x99\x09\x16\xc8\x00\x0d\x60\x80\x46\xa2\xd5\x5a\xc0\x52"
    "\xc5\xa0\x52\x95\x46\x8e\x56\x3a\x7e\x1a\xbe\xc5\xbd\x70\xf7\x8d"
    "\x15\x6a\x19\x4d\x04\xd0\x2d\xb2\xff\xe1\xe3\xa7\xd3\x6d\x80\x4d"
    "\xa6\xcf\x38\xb8\x06\x6f\x3f\x82\x0f\x52\x1c\x6f\x3e\x5d\x82\xe5"
    "\x00\xe5\xf6\x44\x65\xe2\xa8\x83\x12\xa9\xff\x35\xec\x92\xac\x31"
    "\x84\x41\x81\x9d\xf3\x06\xf5\x2b\x2e\xa6\x16\x92\x63\x65\xc7\x70"
    "\x32\x73\x9b\x80\xea\xbd\x20\xac\xb7\xbb\xf7\x2a\xd7\x2d\xf1\x38"
    "\x36\xe7\x4f\xd5\x0b\x2c\x4c\x32\x21\x9c\xa5\xfd\x97\x06\xb6\xf6"
    "\x40\xbe\xe3\x49\x99\x2c\xd3\xbf\x75\xa2\x47\x18\x9d\x37\xa7\x0f"
    "\xd3\xe5\x1d\xc1\x05\xb4\xa0\xa1\xa1\xe6\xfb\x39\x0f\x2f\xa8\x81"
    "\x00\x3e\xcf\xef\x98\x85\x0e\xe2\x62\x00\x78\x66\xc3\xcb\xc0\xbb"
    "\x41\x35\xfd\xa9\xaf\x7e\x98\x94\x60\xa5\xca\x75\xd3\x28\xba\x6d"
    "\xc6\x3c\xb5\x17\x01\xfb\x50\xbe\x9f\xa8\xec\xb0\x41\x16\x97\x3a"
    "\x0d\xdf\x10\xc7\x1f\x2f\x24\xf4\xf3\x2e\xbe\xb3\x5f\x87\x1f\xb2"
    "\xc9\xe8\x06\xb4\x6d\x77\x9d\x3a\x6b\xa1\xad\x12\x87\xf2\x1b\x00"
    "\x09\x15\xda\x52\xa4\x21\xc9\x43\x1d\x03\xe4\x01\xa6\x93\x57\xb5"
    "\x27\x04\xbf\x4a\x39\x84\xc9\xed\x0f\x1f\xf2\xee\xe9\x51\xff\xe6"
    "\xa5\x88\xb1\x0d\xcc\xad\xeb\x33\xfc\x23\xf0\x6b\xe5\xa4\xf5\xcc"
    "\x88\xbc\x76\x5c\x16\x45\x2a\x61\x4f\x14\x9c\xf5\x07\x0c\xe4\x3d"
    "\x38\xc3\x32\x15\x63\xe5\x70\xd2\x87\x22\x0f\x5f\xf3\xa1\xc7\xcd"
    "\xcf\xaf\xaa\x2e\xa1\xed\x86\x98\xcf\x6a\x87\x5f\xdd\xbc\x53\x7d"
    "\x93\x38\x59\x0d\xda\x97\x4c\x01\x84\x5b\xe2\x8b\x4f\x90\x12\x86"
    "\x3b\xf8\x7f\x2d\xbd\x62\x04\x2e\xc1\x93\xc0\x97\x94\x40\xaf\x88"
    "\x47\x3c\xb3\xec\x65\x5c\x5a\x8b\x69\x03\x1f\xc0\x29\x05\xd6\x6a"
    "\x7e\xef\xe3\xe0\xf9\xcd\x0d\x88\x3e\x0b\xee\x9a\x84\x7c\xe4\x45"
    "\xef\x0a\x20\x1b\x27\x63\x2b\x8c\xda\x32\x9a\x17\xad\xc9\xf2\xa8"
    "\x33\xd5\x85\x15\x4a\x71\x65\x79\xa2\xa8\xe3\xed\x69\x38\x10\x54"
    "\x98\x62\xa7\xfc\x93\x68\x6a\x18\xd5\x55\x23\xe1\xf5\x86\x0a\xf9"
    "\xc9\x97\x17\x35\xab\x9e\x2a\x79\x85\xc6\x70\xf4\x0c\xa4\x80\xa9"
    "\x33\x36\x33\xfb\x02\x34\xf2\x6b\x96\xa8\x73\x2a\xa3\x7d\x4c\xb7"
    "\x02\xcf\x58\x15\x2d\x53\x1d\x9a\xcf\x96\xb3\x5e\x05\xce\x29\xb9"
    "\xd7\x7d\xff\xa4\x9c\x56\x07\xe7\x6a\x48\xea\x81\xfb\x2f\x15\x5e"
    "\x8e\xde\x81\x77\x1d\x13\x6a\x9d\xef\xb7\xc5\x7c\x6c\x1a\x73\x16"
    "\x59\x8a\x6f\x24\x85\xf4\xe6\x32\xfa\x9e\xc2\xea\xc2\xcb\xcf\xe5"
    "\xbe\x93\xb8\x52\xc1\xf1\x6c\x05\xd1\x70\x34\xdd\x7b\xe3\x7f\xcf"
    "\x8a\x30\xa8\x1d\xf2\xf9\x25\x09\xa2\x4d\x05\x4b\xba\xa8\x63\x2a"
    "\xb1\xf3\x87\x34\xd8\xdf\x8c\x13\xd8\x72\xb5\x20\x36\x1d\xaf\x7c"
    "\x2a\x8f\xaa\xb4\xd1\xfc\x0b\xc9\xf2\x21\xa2\x07\x8a\xfe\x40\xf0"
    "\x52\x36\x3d\x32\x82\x67\x09\x5d\xf5\x59\xe5\xd2\x39\x09\x91\x49"
    "\xce\x11\x9f\x53\xe1\xbb\x32\x6a\xc5\xf4\x74\x60\xdd\x19\x9e\xf1"
    "\x96\x82\x42\x28\x74\x51\x95\x21\x07\xc0\x8f\xc4\x54\x5f\x12\x66"
    "\x58\x7c\xfe\x24\xb6\xea\x11\xe0\x76\x96\xb1\x25\x45\xa4\xcc\xbf"
    "\x71\x09\x91\xdd\x1e\x47\x8d\x9b\xe8\xcf\xe5\xd3\x19\x87\xaa\x63"
    "\x11\xae\xa5\x50\xa7\xeb\x1f\x6d\x47\x4e\xb8\x7a\x5b\x63\x15\x0f"
    "\x85\x7b\x65\xb1\xd2\x30\xad\x25\x0a\x9c\xb8\xeb\x7a\x62\xb9\x9b"
    "\xb9\x41\x1d\x00\xdd\xb6\x4b\xa0\xa1\xad\xb9\x26\xbd\x7f\xf0\x97"
    "\xd6\x87\xa8\x9c\x57\x40\x9b\xe6\x78\xad\x6f\xbb\xa4\xad\x3e\xb1"
    "\xdd\x3b\x16\x56\x19\xbd\xa5\x31\xf6\xd7\xb5\x2f\xdc\x72\x7f\x2d"
    "\xcb\xff\x02\x09\x8d\x3d\xbb\xcd\x32\x98\xf3\x2a\x0d\x5b\xdc\xd9"
    "\xa4\x99\x59\xab\x70\xa3\xa2\xec\x40\x79\x01\x57\xaa\x38\xb7\xa5"
    "\x76\x92\x92\x1a\x61\x22\xca\x03\x99\x4b\x51\xe0\x01\x6a\xca\x57"
    "\x25\xa6\x56\x4a\x87\xcd\xe8\x9d\x6a\x72\xbd\x2c\x0d\x3c\x30\x5b"
    "\x2c\x52\x60\xbd\x55\xb0\xac\x21\x6a\x98\xda\x5d\xaa\xb6\x57\x91"
    "\xcf\xf0\x81\x3c\xff\xca\x48\x4d\x3a\x94\x19\xfe\xaf\xc1\x55\xba"
    "\x6e\x80\x4d\x49\x24\x29\x81\xbc\x92\xdc\xb2\x89\xa4\x70\x84\x8a"
    "\x2c\xb3\x77\xe3\x73\x10\x47\x51\x0f\x13\x32\x0b\xba\xb5\x70\x88"
    "\x62\x63\x37\x38\xb6\xbb\x4f\xfb\x36\x52\xa8\xc1\xa1\xf0\x8e\x2c"
    "\xad\xce\xf3\x0a\xb0\xfb\xad\x90\x07\x26\xc2\xc1\xeb\x02\xa7\x6d"
    "\xd7\x54\x52\xbe\x0c\x28\xb8\xdf\x13\xd4\xfc\x4a\x37\x36\x8f\x62"
    "\x5a\xde\x5f\x43\xa3\x89\xde\x72\x79\x2d\x5a\x87\x14\x14\x81\x3f"
    "\x07\x2c\xf0\x20\x6a\xf8\xf2\xf2\xb9\x55\x0e\x9a\xda\x03\xa4\x53"
    "\x37\x64\x9c\xa2\xc8\x87\x51\x4a\x25\x10\x9a\x11\x15\xbd\xe6\xea"
    "\x24\xa1\xd9\xfa\x95\x02\x56\x6c\x9b\xe6\x5d\x3f\xdc\xca\x69\x4b"
    "\x3e\x51\xba\x7a\xf1\x71\x92\xd9\x25\x84\x80\x40\x3d\xd2\x15\xf0"
    "\xe5\x65\x3a\xe4\x43\x7a\xd0\x33\xab\xb8\x39\x25\xe0\xf1\x9a\xbf"
    "\x9e\x67\xf5\x73\x74\x70\xab\x98\x9f\x90\xeb\xf4\xfd\x5c\xcd\xb4"
    "\x81\xab\x9a\x85\xf1\x8a\x56\xaa\x10\x30\xd2\x9a\xee\x1b\x60\x1d"
    "\xd1\xfe\x70\xfb\xa6\x02\x94\xbf\x64\x4a\x2e\xad\xa1\x83\x06\x5c"
    "\x71\x00\x88\x28\x85\x14\xbb\x67\x13\xf4\xd7\x15\x5c\x69\xfe\xc6"
    "\x2b\xbe\xdb\xf3\x94\xe4\x95\xeb\x42\xf0\x8c\x21\x25\xd0\x29\x6d"
    "\x24\xc3\x7a\xae\x23\xac\x29\x8d\x15\x03\x36\x87\x3d\x2c\x0d\x98"
    "\x75\xd5\x96\x3c\x41\xc1\x88\x3b\xca\x25\x80\x41\xdc\x2f\x19\x6c"
    "\x9e\x68\x76\x0f\x18\x9d\x2e\x69\xda\x8c\x08\xd2\x7d\x92\x7e\xf9"
    "\x18\xf9\x7e\x2c\xba\xc9\xd6\xcd\x66\x6f\x52\xa6\x06\xca\xd6\x01"
    "\xec\x52\x02\xd1\x97\x2b\xba\xaf\x01\xd0\x90\xe8\xb0\x20\xe9\x53"
    "\xe7\xcb\x37\x35\x4d\x25\xf1\x0f\xb6\x14\x54\x48\x55\xe9\x0d\xfa"
    "\x2e\x66\x39\x07\x18\x3d\xb1\xe4\x87\x29\x64\xb1\x7c\x0c\x47\xf0"
    "\x46\xa9\x8e\xef\x03\x74\x26\x0e\x9b\x5f\x5a\x6f\x97\x21\x55\x35"
    "\x80\x8c\x34\x9c\x9f\xb3\x14\x30\xa8\xb7\xc6\x40\x0d\x4f\xe5\x5b"
    "\x6e\xc9\xe7\xb5\x41\x78\x21\x16\x76\x67\x89\xfe\xd1\xd4\x4b\x5f"
    "\x84\x4d\x96\xc7\x82\xa6\x0d\x65\xc7\x9c\x31\xf1\xc6\x3e\x87\xff"
    "\xc9\xe0\x2c\xf9\x77\x95\x08\x3f\xbd\x00\xfc\x61\xeb\x77\x21\xe6"
    "\x35\x1d\xdb\xd4\xe6\x8b\x6f\xbb\x82\x80\x5d\x58\x9d\x1d\x7a\x59"
    "\x80\x64\xc6\xe2\x66\x7e\x1a\x4a\x4c\x71\xf2\xd8\xe2\xff\x02\x7f"
    "\xc2\x1d\x32\x33\xc6\xe1\xf4\x05\x90\xd1\x55\x1a\x15\x65\xd2\x73"
    "\xdc\x04\x40\x81\xd2\x40\xe3\x59\x35\x9f\x9c\x3c\xdd\x72\x4f\x2d"
    "\xfd\xa8\x9e\x1c\x85\x1c\xac\x3d\xb9\xa0\x2d\x10\xc2\x57\x12\x65"
    "\x51\xda\x80\x03\x34\x52\xd2\xff\xdd\x18\x0f\xbd\x02\x6b\x93\x07"
    "\x64\x88\x13\x2a\x80\x8e\xe5\x8a\xec\x5b\xe9\xae\xf3\x94\x34\x48"
    "\xa5\x88\xb1\x0d\xcc\xad\xeb\x33\xfc\x23\xf0\x6b\xe5\xa4\xf5\xcc"
    "\x88\xbc\x76\x5c\x16\x45\x2a\x61\x4f\x14\x9c\xf5\x07\x0c\xe4\x3d"
    "\x38\xc3\x32\x15\x63\xe5\x70\xd2\x87\x22\x0f\x5f\xf3\xa1\xc7\xcd"
    "\xcf\xaf\xaa\x2e\xa1\xed\x86\x98\xcf\x6a\x87\x5f\xdd\xbc\x53\x7d"
    "\x93\x38\x59\x0d\xda\x97\x4c\x01\x84\x5b\xe2\x8b\x4f\x90\x12\x86"
    "\x3b\xf8\x7f\x2d\xbd\x62\x04\x2e\xc1\x93\xc0\x97\x94\x40\xaf\x88"
    "\x47\x3c\xb3\xec\x65\x5c\x5a\x8b\x69\x03\x1f\xc0\x29\x05\xd6\x6a"
    "\x7e\xef\xe3\xe0\xf9\xcd\x0d\x88\x3e\x0b\xee\x9a\x84\x7c\xe4\x45"
    "\xef\x0a\x20\x1b\x27\x63\x2b\x8c\xda\x32\x9a\x17\xad\xc9\xf2\xa8"
    "\xca\x0e\x78\xab\xde\xea\x5b\xae\x5e\x67\xbd\x26\x24\x29\x3c\x1b"
    "\x8a\xac\x69\x02\x9f\xea\xd3\x2f\x2a\x72\xc3\x60\x2f\xba\xe4\x35"
    "\x77\xb1\x42\x13\x4a\x6b\x18\xe0\x0d\xb3\x2e\xcd\xf5\xee\xf0\x53"
    "\x0f\x1f\x9a\x35\x62\xab\x6b\x1e\x90\xa8\x83\x5b\xb9\x50\x43\x6b"
    "\x8c\x3b\x9b\x83\x57\xe1\x03\x18\x39\xb6\x0e\x6e\x4d\x98\xd7\x4f"
    "\x73\x84\x3e\x10\x9f\x4b\x8f\x5d\xb3\xb8\xff\xaa\x4a\x0b\x4a\xaa"
    "\xf6\xe7\xae\x5a\x98\xf4\xd8\x28\x52\x7d\x53\xf8\xeb\x74\x2b\x2e"
    "\x44\x76\x0d\x4a\x23\x59\xb3\xce\x01\xca\x18\xad\x58\x18\xa1\xe2"
    "\x6f\x3e\x9a\x51\xfa\x8c\xc7\x7c\x45\xfd\xcf\x9e\x93\x20\xfc\x6b"
    "\x68\x89\x03\xe9\xf9\x12\xf2\x78\xbd\xd0\xd7\xb1\x9b\xf9\xc7\x96"
    "\xbb\xe2\x99\x09\x16\xc8\x00\x0d\x60\x80\x46\xa2\xd5\x5a\xc0\x52"
    "\xc5\xa0\x52\x95\x46\x8e\x56\x3a\x7e\x1a\xbe\xc5\xbd\x70\xf7\x8d"
    "\x15\x6a\x19\x4d\x04\xd0\x2d\xb2\xff\xe1\xe3\xa7\xd3\x6d\x80\x4d"
    "\xa6\xcf\x38\xb8\x06\x6f\x3f\x82\x0f\x52\x1c\x6f\x3e\x5d\x82\xe5"
    "\x00\xe5\xf6\x44\x65\xe2\xa8\x83\x12\xa9\xff\x35\xec\x92\xac\x31"
    "\x84\x41\x81\x9d\xf3\x06\xf5\x2b\x2e\xa6\x16\x92\x63\x65\xc7\x70"
    "\x32\x73\x9b\x80\xea\xbd\x20\xac\xb7\xbb\xf7\x2a\xd7\x2d\xf1\x38"
    "\x36\xe7\x4f\xd5\x0b\x2c\x4c\x32\x21\x9c\xa5\xfd\x97\x06\xb6\xf6"
    "\x40\xbe\xe3\x49\x99\x2c\xd3\xbf\x75\xa2\x47\x18\x9d\x37\xa7\x0f"
    "\xd3\xe5\x1d\xc1\x05\xb4\xa0\xa1\xa1\xe6\xfb\x39\x0f\x2f\xa8\x81"
    "\x00\x3e\xcf\xef\x98\x85\x0e\xe2\x62\x00\x78\x66\xc3\xcb\xc0\xbb"
    "\x41\x35\xfd\xa9\xaf\x7e\x98\x94\x60\xa5\xca\x75\xd3\x28\xba\x6d"
    "\xc6\x3c\xb5\x17\x01\xfb\x50\xbe\x9f\xa8\xec\xb0\x41\x16\x97\x3a"
    "\x0d\xdf\x10\xc7\x1f\x2f\x24\xf4\xf3\x2e\xbe\xb3\x5f\x87\x1f\xb2"
    "\xc9\xe8\x06\xb4\x6d\x77\x9d\x3a\x6b\xa1\xad\x12\x87\xf2\x1b\x00"
    "\x09\x15\xda\x52\xa4\x21\xc9\x43\x1d\x03\xe4\x01\xa6\x93\x57\xb5"
    "\x27\x04\xbf\x4a\x39\x84\xc9\xed\x0f\x1f\xf2\xee\xe9\x51\xff\xe6"
    "\xa5\x88\xb1\x0d\xcc\xad\xeb\x33\xfc\x23\xf0\x6b\xe5\xa4\xf5\xcc"
    "\x88\xbc\x76\x5c\x16\x45\x2a\x61\x4f\x14\x9c\xf5\x07\x0c\xe4\x3d"
    "\x38\xc3\x32\x15\x63\xe5\x70\xd2\x87\x22\x0f\x5f\xf3\xa1\xc7\xcd"
    "\xcf\xaf\xaa\x2e\xa1\xed\x86\x98\xcf\x6a\x87\x5f\xdd\xbc\x53\x7d"
    "\x93\x38\x59\x0d\xda\x97\x4c\x01\x84\x5b\xe2\x8b\x4f\x90\x12\x86"
    "\x3b\xf8\x7f\x2d\xbd\x62\x04\x2e\xc1\x93\xc0\x97\x94\x40\xaf\x88"
    "\x47\x3c\xb3\xec\x65\x5c\x5a\x8b\x69\x03\x1f\xc0\x29\x05\xd6\x6a"
    "\x7e\xef\xe3\xe0\xf9\xcd\x0d\x88\x3e\x0b\xee\x9a\x84\x7c\xe4\x45"
    "\xef\x0a\x20\x1b\x27\x63\x2b\x8c\xda\x32\x9a\x17\xad\xc9\xf2\xa8"
    "\x33\xd5\x85\x15\x4a\x71\x65\x79\xa2\xa8\xe3\xed\x69\x38\x10\x54"
    "\x98\x62\xa7\xfc\x93\x68\x6a\x18\xd5\x55\x23\xe1\xf5\x86\x0a\xf9"
    "\xc9\x97\x17\x35\xab\x9e\x2a\x79\x85\xc6\x70\xf4\x0c\xa4\x80\xa9"
    "\x33\x36\x33\xfb\x02\x34\xf2\x6b\x96\xa8\x73\x2a\xa3\x7d\x4c\xb7"
    "\x02\xcf\x58\x15\x2d\x53\x1d\x9a\xcf\x96\xb3\x5e\x05\xce\x29\xb9"
    "\xd7\x7d\xff\xa4\x9c\x56\x07\xe7\x6a\x48\xea\x81\xfb\x2f\x15\x5e"
    "\x8e\xde\x81\x77\x1d\x13\x6a\x9d\xef\xb7\xc5\x7c\x6c\x1a\x73\x16"
    "\x59\x8a\x6f\x24\x85\xf4\xe6\x32\xfa\x9e\xc2\xea\xc2\xcb\xcf\xe5"
    "\xbe\x93\xb8\x52\xc1\xf1\x6c\x05\xd1\x70\x34\xdd\x7b\xe3\x7f\xcf"
    "\x8a\x30\xa8\x1d\xf2\xf9\x25\x09\xa2\x4d\x05\x4b\xba\xa8\x63\x2a"
    "\xb1\xf3\x87\x34\xd8\xdf\x8c\x13\xd8\x72\xb5\x20\x36\x1d\xaf\x7c"
    "\x2a\x8f\xaa\xb4\xd1\xfc\x0b\xc9\xf2\x21\xa2\x07\x8a\xfe\x40\xf0"
    "\x52\x36\x3d\x32\x82\x67\x09\x5d\xf5\x59\xe5\xd2\x39\x09\x91\x49"
    "\xce\x11\x9f\x53\xe1\xbb\x32\x6a\xc5\xf4\x74\x60\xdd\x19\x9e\xf1"
    "\x96\x82\x42\x28\x74\x51\x95\x21\x07\xc0\x8f\xc4\x54\x5f\x12\x66"
    "\x58\x7c\xfe\x24\xb6\xea\x11\xe0\x76\x96\xb1\x25\x45\xa4\xcc\xbf"
    "\x71\x09\x91\xdd\x1e\x47\x8d\x9b\xe8\xcf\xe5\xd3\x19\x87\xaa\x63"
    "\x11\xae\xa5\x50\xa7\xeb\x1f\x6d\x47\x4e\xb8\x7a\x5b\x63\x15\x0f"
    "\x85\x7b\x65\xb1\xd2\x30\xad\x25\x0a\x9c\xb8\xeb\x7a\x62\xb9\x9b"
    "\xb9\x41\x1d\x00\xdd\xb6\x4b\xa0\xa1\xad\xb9\x26\xbd\x7f\xf0\x97"
    "\xd6\x87\xa8\x9c\x57\x40\x9b\xe6\x78\xad\x6f\xbb\xa4\xad\x3e\xb1"
    "\xdd\x3b\x16\x56\x19\xbd\xa5\x31\xf6\xd7\xb5\x2f\xdc\x72\x7f\x2d"
    "\xcb\xff\x02\x09\x8d\x3d\xbb\xcd\x32\x98\xf3\x2a\x0d\x5b\xdc\xd9"
    "\xa4\x99\x59\xab\x70\xa3\xa2\xec\x40\x79\x01\x57\xaa\x38\xb7\xa5"
    "\x76\x92\x92\x1a\x61\x22\xca\x03\x99\x4b\x51\xe0\x01\x6a\xca\x57"
    "\x25\xa6\x56\x4a\x87\xcd\xe8\x9d\x6a\x72\xbd\x2c\x0d\x3c\x30\x5b"
    "\x2c\x52\x60\xbd\x55\xb0\xac\x21\x6a\x98\xda\x5d\xaa\xb6\x57\x91"
    "\xcf\xf0\x81\x3c\xff\xca\x48\x4d\x3a\x94\x19\xfe\xaf\xc1\x55\xba"
    "\x6e\x80\x4d\x49\x24\x29\x81\xbc\x92\xdc\xb2\x89\xa4\x70\x84\x8a"
    "\x2c\xb3\x77\xe3\x73\x10\x47\x51\x0f\x13\x32\x0b\xba\xb5\x70\x88"
    "\x62\x63\x37\x38\xb6\xbb\x4f\xfb\x36\x52\xa8\xc1\xa1\xf0\x8e\x2c"
    "\xad\xce\xf3\x0a\xb0\xfb\xad\x90\x07\x26\xc2\xc1\xeb\x02\xa7\x6d"
    "\xd7\x54\x52\xbe\x0c\x28\xb8\xdf\x13\xd4\xfc\x4a\x37\x36\x8f\x62"
    "\x5a\xde\x5f\x43\xa3\x89\xde\x72\x79\x2d\x5a\x87\x14\x14\x81\x3f"
    "\x07\x2c\xf0\x20\x6a\xf8\xf2\xf2\xb9\x55\x0e\x9a\xda\x03\xa4\x53"
    "\x37\x64\x9c\xa2\xc8\x87\x51\x4a\x25\x10\x9a\x11\x15\xbd\xe6\xea"
    "\x24\xa1\xd9\xfa\x95\x02\x56\x6c\x9b\xe6\x5d\x3f\xdc\xca\x69\x4b"
    "\x3e\x51\xba\x7a\xf1\x71\x92\xd9\x25\x84\x80\x40\x3d\xd2\x15\xf0"
    "\xe5\x65\x3a\xe4\x43\x7a\xd0\x33\xab\xb8\x39\x25\xe0\xf1\x9a\xbf"
    "\x9e\x67\xf5\x73\x74\x70\xab\x98\x9f\x90\xeb\xf4\xfd\x5c\xcd\xb4"
    "\x81\xab\x9a\x85\xf1\x8a\x56\xaa\x10\x30\xd2\x9a\xee\x1b\x60\x1d"
    "\xd1\xfe\x70\xfb\xa6\x02\x94\xbf\x64\x4a\x2e\xad\xa1\x83\x06\x5c"
    "\x71\x00\x88\x28\x85\x14\xbb\x67\x13\xf4\xd7\x15\x5c\x69\xfe\xc6"
    "\x2b\xbe\xdb\xf3\x94\xe4\x95\xeb\x42\xf0\x8c\x21\x25\xd0\x29\x6d"
    "\x24\xc3\x7a\xae\x23\xac\x29\x8d\x15\x03\x36\x87\x3d\x2c\x0d\x98"
    "\x75\xd5\x96\x3c\x41\xc1\x88\x3b\xca\x25\x80\x41\xdc\x2f\x19\x6c"
    "\x9e\x68\x76\x0f\x18\x9d\x2e\x69\xda\x8c\x08\xd2\x7d\x92\x7e\xf9"
    "\x18\xf9\x7e\x2c\xba\xc9\xd6\xcd\x66\x6f\x52\xa6\x06\xca\xd6\x01"
    "\xec\x52\x02\xd1\x97\x2b\xba\xaf\x01\xd0\x90\xe8\xb0\x20\xe9\x53"
    "\xe7\xcb\x37\x35\x4d\x25\xf1\x0f\xb6\x14\x54\x48\x55\xe9\x0d\xfa"
    "\x2e\x66\x39\x07\x18\x3d\xb1\xe4\x87\x29\x64\xb1\x7c\x0c\x47\xf0"
    "\x46\xa9\x8e\xef\x03\x74\x26\x0e\x9b\x5f\x5a\x6f\x97\x21\x55\x35"
    "\x80\x8c\x34\x9c\x9f\xb3\x14\x30\xa8\xb7\xc6\x40\x0d\x4f\xe5\x5b"
    "\x6e\xc9\xe7\xb5\x41\x78\x21\x16\x76\x67\x89\xfe\xd1\xd4\x4b\x5f"
    "\x84\x4d\x96\xc7\x82\xa6\x0d\x65\xc7\x9c\x31\xf1\xc6\x3e\x87\xff"
    "\xc9\xe0\x2c\xf9\x77\x95\x08\x3f\xbd\x00\xfc\x61\xeb\x77\x21\xe6"
    "\x35\x1d\xdb\xd4\xe6\x8b\x6f\xbb\x82\x80\x5d\x58\x9d\x1d\x7a\x59"
    "\x80\x64\xc6\xe2\x66\x7e\x1a\x4a\x4c\x71\xf2\xd8\xe2\xff\x02\x7f"
    "\xc2\x1d\x32\x33\xc6\xe1\xf4\x05\x90\xd1\x55\x1a\x15\x65\xd2\x73"
    "\xdc\x04\x40\x81\xd2\x40\xe3\x59\x35\x9f\x9c\x3c\xdd\x72\x4f\x2d"
    "\xfd\xa8\x9e\x1c\x85\x1c\xac\x3d\xb9\xa0\x2d\x10\xc2\x57\x12\x65"
    "\x51\xda\x80\x03\x34\x52\xd2\xff\xdd\x18\x0f\xbd\x02\x6b\x93\x07"
    "\x64\x88\x13\x2a\x80\x8e\xe5\x8a\xec\x5b\xe9\xae\xf3\x94\x34\x48"
    "\xa5\x88\xb1\x0d\xcc\xad\xeb\x33\xfc\x23\xf0\x6b\xe5\xa4\xf5\xcc"
    "\x88\xbc\x76\x5c\x16\x45\x2a\x61\x4f\x14\x9c\xf5\x07\x0c\xe4\x3d"
    "\x38\xc3\x32\x15\x63\xe5\x70\xd2\x87\x22\x0f\x5f\xf3\xa1\xc7\xcd"
    "\xcf\xaf\xaa\x2e\xa1\xed\x86\x98\xcf\x6a\x87\x5f\xdd\xbc\x53\x7d"
    "\x93\x38\x59\x0d\xda\x97\x4c\x01\x84\x5b\xe2\x8b\x4f\x90\x12\x86"
    "\x3b\xf8\x7f\x2d\xbd\x62\x04\x2e\xc1\x93\xc0\x97\x94\x40\xaf\x88"
    "\x47\x3c\xb3\xec\x65\x5c\x5a\x8b\x69\x03\x1f\xc0\x29\x05\xd6\x6a"
    "\x7e\xef\xe3\xe0\xf9\xcd\x0d\x88\x3e\x0b\xee\x9a\x84\x7c\xe4\x45"
    "\xef\x0a\x20\x1b\x27\x63\x2b\x8c\xda\x32\x9a\x17\xad\xc9\xf2\xa8"
    "\x8d\x5e\x97\x32\x66\x4f\xab\x7a\xc3\x55\xf1\x6c\xb6\xf4\x3f\xc2"
    "\xf9\xd5\xbe\xfd\x07\x83\x8d\x0f\x9f\xf5\xa0\x88\x13\x71\x53\xc8"
    "\x37\xa5\x6d\xb5\x84\xce\xee\xb3\x03\x9b\xdc\x86\x24\xe9\x41\x68"
    "\xa5\x44\xab\x59\x69\xa9\x85\xeb\x60\x8a\x9b\x73\x4a\x46\x84\x68"
    "\x70\x36\xbb\x77\x0d\xfc\xa0\x03\xc1\xbc\xa8\xdd\xde\xaa\x53\xab"
    "\xd0\x34\x15\x6e\xcf\x1c\x57\x2d\xa8\xaf\x14\xf0\x0d\x07\x2d\x76"
    "\x03\xfd\xbc\x41\xf4\x18\x26\xe1\x11\x32\x79\x57\x0c\xa5\x55\xd4"
    "\x20\x3b\xa0\x7b\x76\xa9\x62\xe0\x6f\x5b\xb8\xf0\x59\x06\x25\x4a"
    "\xdd\x7c\x15\x08\x1f\x69\x0a\xa2\x20\x84\x9b\x58\x97\x58\x70\x61"
    "\x66\x1d\x79\x00\xc8\xad\x08\x50\xbf\xeb\xd8\x12\xb2\xf6\xbb\x06"
    "\xac\xa8\x76\x49\xc5\x69\xae\xc3\x99\x0e\x4d\xb0\x27\x90\x5f\x8c"
    "\x2e\xab\xe3\x3b\xb5\xfc\x36\x51\x83\x71\x2f\x1c\x54\x95\xa2\xef"
    "\x13\xf1\xe4\x71\x66\x4c\xce\x81\x95\x98\x96\x10\x3a\xde\xe1\x55"
    "\xaa\x2d\x65\x3f\xf8\xf3\xe2\x93\x6a\xc7\x6e\xba\x4d\xcd\xbd\xc1"
    "\x62\xf4\xb3\x6b\x70\xdd\xf2\x21\x61\x9e\xb4\x10\x52\x2e\xef\xa9"
    "\x13\x97\x47\x86\x61\x9a\x25\xa7\x1c\xc8\xc0\x52\xfb\x75\x18\xda"
    "\x2a\x4e\x5d\xfc\xc2\x63\xfd\x5c\x2a\x54\x30\x42\x51\xce\x64\x14"
    "\xd6\xe1\xae\x61\xd0\x06\xdb\xe1\x51\x6d\x03\x9b\x2f\x34\xf7\x87"
    "\x91\x0e\x0b\xca\xab\x1c\x91\xdc\x2a\xb2\x48\x93\x38\x61\x8f\x7f"
    "\xef\x2f\x05\xcc\xc1\xc5\x2c\x3c\x35\xb1\x9d\xc3\x05\x73\xe8\xb4"
    "\x0e\x73\x3d\x0d\x8b\x53\x09\x03\x27\xb5\x8d\xfe\x5f\xe8\x87\x46"
    "\x46\xfa\xcb\x5f\x1a\xe7\x6b\xda\x8a\xd7\x59\x4f\x0c\x45\xad\x68"
    "\x44\x4c\xd8\x0b\x75\x3f\x7e\x47\x6c\xe2\x8d\xcc\xb6\x5b\x54\x84"
    "\x64\x22\xb7\x46\x67\x14\xfd\x77\xbe\xec\xc1\x5d\x06\x9c\x16\x26"
    "\x09\x57\x99\x6a\x01\x7e\xac\x51\xb1\x61\x44\x5c\xf1\x11\xf9\x6e"
    "\x46\xd4\xea\x36\xcf\x7b\xf8\x2c\x13\x6b\xba\xd4\x20\x3b\x81\x59"
    "\x4d\x4d\x92\xf3\x24\x95\x0a\x75\x55\x50\x70\x58\x74\x04\x78\x5a"
    "\xd0\x14\xda\xc3\xd9\x38\x10\xfd\x70\x86\xc2\xa0\x6d\x98\x69\xdc"
    "\x53\x0a\x5a\x2f\x92\xcf\x4a\x21\x3c\x2d\xcd\xa9\xc7\x8e\x17\xa7"
    "\x5c\x1f\xb3\x06\x6a\xa4\xad\xb7\x47\x92\x98\x25\xb3\x7b\x25\xef"
    "\x7f\x7f\x45\x56\x8c\xf4\x55\xe1\x17\x09\x34\xf9\x8c\x69\x06\x91"
    "\x5c\xde\xca\xcb\xae\xa4\x0b\xd7\x53\x4f\xb5\x74\x53\xd2\xf1\xfb"
    "\xd4\xa2\x3c\xda\xf1\xd5\xda\x77\xbb\x33\x75\x06\x8a\xf6\x41\xe2"
    "\x20\xff\x59\xe3\x9c\xf2\x74\xa7\x90\x2d\xa2\x01\x47\x5c\x8a\x58"
    "\x83\xfe\xdd\x86\xa9\x7c\x86\xf5\xf3\xd5\x3a\x3f\xcc\x69\x72\x76"
    "\xc6\x84\x5b\xec\x6b\xc4\xb1\xed\x50\x8c\x5c\x09\x12\x6e\x92\x0d"
    "\xb9\x65\xf1\x02\x34\x27\x2b\xe8\x88\x2e\x89\x0e\xdb\xac\x17\x52"
    "\x5a\x36\x1d\xe7\xea\xff\x72\xeb\x1f\xd4\x78\x7d\xc4\xeb\x59\x32"
    "\x32\x9f\x09\xa8\xd2\x19\xfa\x11\x7f\xf8\x68\x09\x46\x68\x57\x77"
    "\x64\x86\x1e\x5c\xef\x44\xae\x05\x33\x32\x83\xd1\xbc\xce\x3f\x5f"
    "\xff\x16\x6d\xfb\x44\xea\xb6\xfc\xf3\xff\x22\xc3\x61\x58\xe0\xb8"
    "\x64\xd0\x48\xba\x17\x93\x85\xd5\xa5\x3a\x22\x96\xe1\xc3\x55\x6f"
    "\x6e\xf4\xc9\x5d\x01\x7f\xe0\xd6\x92\xd7\x8d\x03\x30\x3d\x44\x7a"
    "\x6c\xec\x48\xb6\xf0\xdc\xac\xa9\x88\x6f\x4b\x10\xdf\x31\x47\x86"
    "\xe8\xf3\xf1\xa5\x92\x73\xa2\xb6\xbf\x55\x6c\xc5\xbc\xb9\xf0\x05"
    "\xab\xfe\xf0\x76\xe0\xc8\x96\x79\xd0\x07\x05\xfc\x52\xe3\xef\x89"
    "\x8f\x22\xbe\xee\x94\x50\x28\xb0\xf1\x9b\xc2\x7a\xc2\x28\x0f\xd8"
    "\xd7\xa9\xbb\x8b\x99\x6e\xd5\x68\x8c\x9b\x0f\x08\xaf\xf2\x20\xc1"
    "\xf0\x2b\xd4\x1f\x4c\x26\x04\xa5\x55\x8d\xb9\x9d\x42\x1b\x0e\xf9"
    "\xca\xc3\x9a\x6e\x63\x24\x97\xde\x8b\xdf\x92\x9e\xa9\x29\x85\x91"
    "\xd7\xd0\x8b\x42\x68\xca\x86\xe7\x65\x58\x02\x66\x3d\x62\x2b\x38"
    "\x2a\x3c\x93\x70\xdc\xfd\xdf\x3e\xd7\x78\x83\x43\x48\x9d\x9c\x3e"
    "\x8f\xf7\x17\x9c\xe1\xe2\x06\x81\x5e\xad\xa9\xd5\x4c\x68\xed\xc9"
    "\xf2\x75\xcb\x27\x07\x00\x50\x96\xa6\xa9\xff\xad\x74\xda\xb1\xab"
    "\x41\xbb\xd8\xba\xb0\xcc\xf7\x3f\xc3\x10\xdb\xa7\x63\xf6\x26\x01"
    "\x79\x54\x08\x73\x7c\x3f\xae\xec\x65\xbb\x10\x87\x71\x16\x06\x2a"
    "\xf6\x1d\xf6\x67\x1b\x68\x6a\x97\x57\x4f\x46\x94\x08\x34\x68\x53"
    "\x1a\xfd\x5a\xb8\x7b\xa9\x4a\xfa\x5e\x42\xe4\x74\x20\x6a\xb7\x3f"
    "\x27\x3c\xc2\x04\xa5\x6e\x89\x9e\x13\x7a\x3f\x62\xf8\xfe\xa1\x57"
    "\x9c\x0d\x43\x25\x29\x69\x88\x98\x23\x48\x4e\x8a\x4a\x11\x61\x88"
    "\xa6\x2c\x97\x8a\x75\x8d\x8f\x6e\x29\xe7\x4b\x5d\x4e\x80\xc0\xff"
    "\xc3\xbd\xe1\x68\xb6\x1e\x13\xab\xd9\x72\x4b\x32\x7d\x76\xa3\xa6"
    "\xa7\x0b\xe3\xaf\x3c\x7d\x35\x83\x46\x2c\x99\x86\xe5\xee\x57\xa7"
    "\x7e\x14\x3d\x42\xb7\x1f\x6f\x22\x0e\x61\x6f\x92\x7d\x73\xef\xd0"
    "\x29\x53\x64\x50\xf0\xfc\x49\x8f\x1a\xc2\xaf\x44\x0b\x99\xc8\x7d"
    "\x55\xb0\xf3\xaf\x18\xc8\x18\x00\x29\x67\x27\xee\x84\x59\x87\xf3"
    "\x69\x1d\x0b\xcc\x08\x78\xe0\xe4\xb7\xb1\x5f\x29\xc5\x4f\xaa\x9d"
    "\x37\xe6\x79\x6d\xfb\x57\xe2\xc3\xd3\x7a\xe9\x26\x92\x25\xc8\x8b"
    "\x97\x98\x0d\x9b\xa2\x64\x85\x37\xa2\x32\x14\xfe\x28\xa7\x61\x06"
    "\xd8\xe6\x1a\x80\xa1\xcf\x97\xad\xcd\xe1\x34\xdb\xa4\x4b\xec\x40"
    "\x2f\xa2\xab\xca\xdb\xd5\x0a\xed\x7d\xe2\x72\x24\x9f\x76\xf5\x02"
    "\xda\xfa\xec\x2d\x39\x10\xa0\xed\x2e\x12\xa9\x40\x62\xa4\xfd\x3c"
    "\x85\x93\x98\x9d\x90\x3b\xcb\xe1\x1f\xf9\xcf\x3b\x8a\x30\x4a\x8e"
    "\xdd\xb1\x26\x40\x04\x71\x64\x7b\xf4\xa8\x10\xce\x05\xbf\x99\x93"
    "\x68\x0d\x2b\x56\x51\xe8\xbb\x9f\x96\x98\x65\x3f\x7b\xa0\x04\x27"
    "\xcb\xbe\x75\xaf\x5c\x7d\x35\x78\x6c\x52\xb3\x43\x38\xed\x8a\xa0"
    "\xea\x28\x12\xf5\x19\xd2\x04\x1f\xc1\xb0\xff\xbc\x96\x52\x8c\x4c"
    "\x58\x88\x91\x1b\x3d\x06\xf6\x83\x43\x02\x71\x38\xba\x1b\xcd\x6f"
    "\x7c\xf2\x09\xb1\xc5\x71\x21\x3b\xe5\xe6\x04\xae\xf9\x7c\xef\x8c"
    "\xa8\xf6\x9b\xc1\x06\xdc\x85\x31\x7a\x58\x40\x1a\x32\x9e\xd5\xaf"
    "\x9f\x96\x3f\x37\x73\x15\xc9\x48\x48\xf9\x53\x5d\x80\x46\xa4\x03"
    "\x67\xd1\xeb\xc8\xb8\xec\x0e\xd9\x00\x33\x1a\x44\x9d\x27\xae\x55"
    "\x9a\x2f\x1b\x29\x83\xd7\x67\xf2\x12\x88\x39\xbc\x29\xe4\x66\xac"
    "\x8d\xc3\xe0\x67\xbb\xe0\x77\x4e\xd1\x12\x43\x86\x31\x8c\xf9\x49"
    "\x07\xdc\xb6\x22\x89\xd1\x17\x77\x10\xa3\x6b\xc9\x8a\x69\x3e\xcd"
    "\x7e\x56\x3e\xe1\x5f\x1f\x8e\x58\x58\xc0\xd9\x87\xce\x23\x7f\x03"
    "\xaf\x9d\x81\x8c\x2c\x73\x49\xa8\x26\x4e\x82\x0a\x46\xbe\xfc\x80"
    "\x39\xf2\xec\x28\xff\x0d\x67\x4d\x9f\xd9\xfd\x7f\x8d\x55\x69\x5f"
    "\x5e\x0d\x11\xa2\x8e\x36\x00\x16\xaa\x3a\x58\xc2\x36\x09\xd5\xe2"
    "\x75\x3a\xb1\xc2\xfb\xe5\x4d\x7e\xa1\xcd\x05\xbb\xd8\x92\x67\x45"
    "\x6b\x4d\x79\xec\x6a\x9a\xa7\x00\xeb\x1c\xa2\xfe\xa4\x96\x27\x8c"
    "\x9d\xdc\xc8\xbb\xee\x33\x18\x3b\x81\xd5\x95\x02\x8c\x82\x88\x21"
    "\x28\x56\x12\x98\xb6\x11\xb8\xa5\x64\x1c\xe6\x10\x4a\xd8\xe1\xc4"
    "\xbe\xd0\x35\xb2\x69\x87\x86\x43\xac\xf4\x20\x5a\x0f\xbd\xd4\x31"
    "\x28\x62\x71\xf0\x6d\x08\x24\x7b\xee\xcf\xdf\xdd\x03\x83\x48\xca"
    "\x8a\xb3\xb9\xc3\x04\x0c\xea\xa5\xe6\xaf\x17\x38\x3c\xc5\x31\x33"
    "\x67\xa3\xa3\xd6\xcc\xc9\xe0\xfb\xc2\x90\xc1\x17\xa0\x43\x39\x67"
    "\x78\xc2\x51\x97\x31\x1f\x85\x6e\x83\x49\xa7\xb5\xcf\x7f\x69\x1f"
    "\x99\x97\x44\x48\x08\xfc\x44\xa3\x3a\x88\x62\x72\x2a\x15\x1c\x62"
    "\x60\x9f\x8e\x01\xec\xce\xa4\xed\xe9\xbf\x0f\x52\x93\xa0\x7a\x80"
    "\x63\x4b\x8f\x9e\x12\x3c\x58\x7a\x64\x1a\x72\x17\xa8\x74\x44\xe2"
    "\x8f\x57\xb9\x53\xd0\x00\xfe\x5c\xa2\xa7\xfc\xcf\x89\xe3\xb6\xa6"
    "\x88\x83\x84\x59\xb6\x36\x21\x8e\xf9\xc0\x0b\x34\xc3\x9c\x44\x93"
    "\x15\xd7\x7a\x06\xa4\x28\x84\xcb\xca\x06\x90\x4f\x5c\x5b\x9f\x19"
    "\x09\x4e\xed\xdd\x50\x8b\xab\x62\x05\x24\xeb\x8f\x1a\x30\x42\xad"
    "\xd5\x55\xbf\xac\x52\xe3\xef\x24\x34\xad\x2c\xc2\x35\xb5\xe2\x07"
    "\x1c\xd0\xf5\xe0\xae\x78\xfb\xbb\x84\x0f\xeb\xe3\x8f\x44\xea\xd2"
    "\x33\x75\x1e\x94\xb6\x83\x7d\xe8\x11\xab\x3a\x93\x8c\x81\x6f\x61"
    "\xeb\xdb\x91\x1d\xe5\x98\x05\x17\x13\xcb\xef\x67\xfc\x4a\xbf\x5f"
    "\x1d\xb4\xed\xe9\xc5\x39\xe9\xd6\x77\x93\x99\x39\xc3\x30\x16\x08"
    "\x56\x54\xe8\xb8\xd9\x1e\xa4\x04\x1e\x56\xb2\x2c\xef\x4e\xd4\xa1"
    "\x4d\xba\xf1\xad\x23\x12\x45\x9c\xf6\xae\x84\xc3\xfe\x49\xd5\xba"
    "\xc8\xc3\xde\xf8\x20\xfc\xf0\x1b\x84\xb0\x5c\x9e\x27\x48\x46\x0a"
    "\x94\x59\xfe\x8f\x3b\xe0\x5d\x81\xb3\xc9\x47\xe8\xa1\x59\x0d\x27"
    "\x11\x0d\x04\xec\xf0\x0a\x62\x3b\xbe\xbd\x99\x9d\x86\x6e\xfc\xe0"
    "\x19\x7b\xf1\x8f\xc4\x64\xdd\x86\x51\x6d\xcf\x3b\x1e\x52\xa7\x30"
    "\xd1\x8d\xe0\xed\x3f\x4c\x27\x64\x9c\x07\x37\xdb\x03\xb3\x47\x62"
    "\x03\x95\x19\x9c\x35\x13\x34\xf5\x09\xc0\x99\x97\xd2\x52\x59\xbe"
    "\x3c\xb5\xb7\xba\x4b\x63\xe5\x0c\xf1\x19\x5a\x16\xed\x63\x3e\x15"
    "\xbd\x3f\xc5\x6f\x27\xed\x7d\x4f\x22\x0a\x43\x59\xab\x31\x3c\xdf"
    "\x8c\xb9\x77\x84\xab\x35\xdb\x92\xd6\xaa\xc0\x3c\xf0\x9f\xc4\x02"
    "\x5a\xbf\x10\x75\x5d\xda\xe9\xe3\x55\xc9\x63\x4a\xfc\xee\x5e\xeb"
    "\x6e\x05\x41\x02\xa9\x61\xbd\x4b\x08\x8c\x4d\x98\x9a\x84\x25\x13"
    "\x98\xd2\x85\x90\x24\xcc\x39\x30\x4b\x06\x8c\xf0\x28\x8d\xf9\x50"
    "\x56\x3f\x8d\xa3\x2e\xd1\x3e\x63\x08\x21\xad\xb0\x66\x78\xa0\xeb"
    "\x29\xf9\xc7\x13\x4e\xe7\x0c\x6f\xaf\xd9\x05\x76\x7c\xcf\x59\x84"
    "\x19\x3d\x5c\x35\x92\xa7\xe2\x2e\xfb\x90\x1a\xbd\xf8\x26\x7d\xcb"
    "\xba\xdb\x8b\x98\xb8\x6b\xd0\x95\xc2\x4e\x41\xe1\x7e\x68\xfd\x4c"
    "\xa2\x06\x91\x65\x55\x29\xfd\x29\x80\x90\xfa\xd6\xf4\xc1\x28\x80"
    "\x68\xfc\xe3\x23\xf2\x0a\xca\x71\x35\x16\xfa\xf9\x64\x3a\x50\x79"
    "\x1b\x20\x38\x05\x56\xf2\x3a\x96\x99\x46\xb4\x2f\xbf\x1c\x98\x69"
    "\x4d\x66\x55\xdc\x70\xbf\x0b\xca\x51\x91\x20\x09\xc5\x51\xbb\x1b"
    "\x4e\x9d\xab\x65\x1b\x13\x79\x8d\x2a\x37\x5a\x68\xdc\xdc\xa2\x96"
    "\x10\x98\xee\x08\x91\x87\x6f\x52\x63\x6b\x0e\x33\x0f\xeb\xfb\x90"
    "\x95\x4e\x5f\x86\xb9\x22\x40\x6f\x38\x56\xb7\x57\x65\x52\xad\x0d"
    "\xa9\xc9\x8f\x48\xe4\x44\xd2\xc0\x92\x80\x8a\x19\x7e\x04\x60\x2b"
    "\x04\x3b\xd2\x05\x18\xef\x0b\xd5\xb9\x79\xed\x17\x6b\xf8\x4c\x1b"
    "\xae\x01\x2e\xa6\x2b\xf1\x93\x9b\x05\xa2\x94\xbe\x81\x80\x24\x3f"
    "\xf1\x28\xd6\xdf\xe8\x48\xaa\x25\xab\x25\xbb\x55\x99\xfc\x4f\x82"
    "\x8f\xfd\x5e\x40\xe7\x52\x9c\xd6\xc1\xd0\x3d\x46\xdb\xde\xe7\x12"
    "\x0b\x10\xea\x88\xd8\x00\xa0\xcc\xee\x4b\x54\x6b\x77\x9d\x5d\xad"
    "\xee\x1d\x68\x9d\x68\xe3\xc9\x73\xd4\x14\x66\x0e\x82\x1d\x98\x1b"
    "\x4d\x6c\xe2\x53\xf9\xcc\x1c\x16\x87\x34\xc8\xe2\x60\x72\x58\x04"
    "\x36\x29\xdb\xe8\x59\x3c\x43\x7f\x7a\xca\xfa\x96\x14\x51\x11\xeb"
    "\x86\xc7\x3f\x7e\x53\x6f\x32\x45\x14\x8d\xb4\x4f\xcb\xf4\xb2\x26"
    "\x83\xc3\x3d\x59\x2d\x22\x3f\x14\x9a\x62\x67\xcf\xc7\xa8\xde\xbe"
    "\x91\x6d\xc9\x4d\xb1\xe6\xa0\x76\x4c\xda\x67\x09\xd8\xd5\xec\x22"
    "\xe5\x0b\x99\x20\x5c\x7c\xd9\x73\x40\xf3\x4d\x32\x90\xbd\xe1\x9a"
    "\x08\xd2\x3d\x9a\x15\xb0\xb9\x25\x56\x47\x63\x60\xea\xe8\x1c\xfb"
    "\x92\xe8\xfb\x2c\x2e\x34\xdb\x1d\xe5\x0e\x34\x05\xc4\x82\x4e\x5f"
    "\xeb\xfd\x85\xf1\xda\x9d\x07\xf5\x80\x6d\xeb\x42\xee\xbc\x9d\x71"
    "\xbd\xb7\x5a\xb8\x07\x56\x30\xb2\x49\x5d\xae\xa4\x3a\xf8\x70\x2a"
    "\xdc\x54\xbb\x56\x00\x68\x10\x53\xc2\x9b\x91\xfa\xf6\x34\x6e\xcc"
    "\x4b\xce\x96\xed\x56\xac\x10\x84\x89\x3d\x41\x99\x84\xda\xc5\xa7"
    "\x9f\x0f\xd7\x28\x3a\xad\x51\x31\x9e\x15\xe0\xd4\x5d\x95\x11\xc1"
    "\x3f\x8b\x54\xf5\x80\x81\x5d\x1e\xe9\xe4\xa0\x3e\xa3\xa7\x1f\xf4"
    "\x28\xc1\x7b\x20\xba\x08\xc6\x8c\x01\xdb\x48\xde\x28\x58\x5c\x60"
    "\xa2\xc6\x6b\x72\x8b\x35\xae\xc8\xa7\x84\xa4\x8b\x1c\x10\x3e\x58"
    "\x6f\x1b\xd9\x9d\xe8\x53\x10\x64\x79\xce\x3f\xf7\xae\x4c\x66\x9c"
    "\xd2\x9a\x4f\x25\xe5\x77\xca\x05\xe4\x98\x3d\xc0\x47\x53\x66\x56"
    "\x0b\xc0\x39\x35\x83\xbd\x0c\x83\x47\xb7\x4b\xf6\xaa\xa7\x42\x10"
    "\x58\xb8\x40\x22\xec\x8c\x0f\x03\x60\x8e\x9f\x43\x1e\xff\x63\xd0"
    "\x4b\xea\x42\x5b\x20\xaf\x91\x23\x5d\x78\xa3\x17\x8c\x10\x8d\x7c"
    "\xbc\x25\xe1\x08\xd7\x96\x0a\x1f\x96\x9a\x03\x0e\x50\xa8\xc3\xed"
    "\x0a\xe5\x9d\x57\x96\xd3\xc6\xfc\xe2\x32\xeb\x7a\xf0\xb4\xf3\x12"
    "\xc8\x74\x74\x6c\x48\xcc\xa6\xa6\x7a\xa5\xa8\x67\xfb\xcb\xc5\xb0"
    "\x5d\x69\x79\x88\xa3\x37\xba\x93\x90\x54\x3d\x37\x74\xd4\x61\xfc"
    "\x0a\xc5\x50\x9f\xab\x75\x6f\x63\xcc\x5b\x84\x80\x65\xff\x1c\xb4"
    "\x8f\x58\x63\xc0\x4f\x90\x53\x88\x2c\x60\x79\x6d\x89\xd1\x40\xd2"
    "\x06\x89\x8d\xda\xc1\x12\x90\x3b\xb4\xf9\x03\x0d\xc5\x2c\x85\x73"
    "\x6e\x7a\xec\x22\x3a\x99\x9f\x1b\xb5\x44\x44\x58\x2c\x03\x41\xfc"
    "\xbd\x9a\xc5\xe9\xc6\x53\xc6\x8d\x8a\x1d\x8e\xf1\x66\xb9\x1e\xd6"
    "\xe4\x83\xc2\x0c\x72\xc8\x96\x6e\x2d\xba\xb2\xec\xd3\x4d\xaf\x86"
    "\xf9\x51\xd7\xc3\xca\x0b\x8c\x58\xc6\x3c\x7b\x6e\xab\x93\x7f\x41"
    "\xf4\x33\x3b\x10\xc7\xdb\x90\x77\x84\x91\x6b\x03\x21\x21\x3c\x28"
    "\xf4\x29\xfa\x47\xe5\x5f\x4b\xef\xc8\xa8\xa2\xb0\x0a\x16\xaf\x5f"
    "\xed\xe8\x36\xc9\xf0\x06\x01\x64\xce\x1f\x72\x9e\xa8\xd9\xd1\x51"
    "\x42\x16\x0d\x85\x46\xbe\x92\x86\xd7\x6e\xcc\xce\x49\xbb\x55\x0f"
    "\x53\x1b\xc8\x5f\x37\xe5\x50\x85\x6c\x56\x89\xc0\x41\x65\x37\x3e"
    "\x9a\x79\x2c\x9e\x69\x47\x4a\x35\x3f\xab\x98\xb1\x86\xe5\x19\x3c"
    "\xa2\x2f\x2d\xc5\x51\x07\x1a\xcd\xa2\xfe\xbf\xeb\xed\x5b\x09\xfc"
    "\xff\x2f\xe3\xb6\x6b\xbe\xab\x19\x22\x4f\xea\xe4\xf4\x4e\x69\xeb"
    "\x56\x8e\x06\x5a\x25\xa9\xe6\x7b\x8d\x9d\xf9\xf4\x5b\x54\x22\xce"
    "\x26\x81\x23\x84\x63\x09\x3f\x4a\x8f\x61\x89\x37\x4d\x86\x5d\x78"
    "\x0c\x40\x7f\x36\x85\xa0\x6c\x66\x68\xcb\x30\x2f\x0f\x12\xa7\x40"
    "\x49\xd1\xe9\x4d\x46\x75\x70\xf1\xe7\xfd\x4b\x93\x2a\x95\xfa\x1c"
    "\x4c\x20\x26\x42\x42\x8d\xb5\x90\x37\x1b\xf7\x2a\x29\x5c\xe4\xdb"
    "\x05\xe0\x03\x5d\x89\x86\x9d\xdb\xc8\xb9\xd1\x94\xae\x40\x01\x76"
    "\x4c\xd2\xe0\x16\xe9\xdc\x26\xbb\xd9\x89\x44\x8e\x1f\x03\xf6\xd7"
    "\x66\xef\x87\x8a\xc1\x75\x5d\xf7\xa2\x18\xe9\x38\xc2\x2a\x75\x94"
    "\x07\xa6\x31\x7d\xcb\xd3\x17\xbe\x89\xc9\x81\x78\x39\x90\xde\x6d"
    "\x03\x14\x01\xe3\x2d\xf9\x1b\xc0\x19\x7f\x3d\xe3\xf7\x5d\xad\x68"
    "\x7f\x69\x79\x4d\x59\x92\x44\xd3\xd3\x12\xfd\x86\x80\x98\xd0\x4e"
    "\x8a\x33\x4c\x17\x3d\xe4\xa1\x67\x8a\x4d\x2d\xeb\x0c\xb6\x49\x06"
    "\xef\x24\x57\xba\xe7\x0e\xfa\x05\xd1\xd3\x68\xa0\xa1\x3a\xb6\xa9"
    "\x0d\xb7\x38\x05\x47\x8f\x4c\x20\xb2\xc1\x8c\x0a\x3a\x5c\x1b\xf8"
    "\x1b\x27\x00\xa4\xb6\x76\x2a\x09\x21\x34\xc1\x30\x16\x0a\x8e\xf4"
    "\x12\x8b\x32\x9d\x39\x93\xda\x57\x7e\xdb\x53\x24\x8d\x63\x82\x86"
    "\x5b\x94\x8d\x3a\xa6\xb9\x3b\x2b\x49\xeb\x61\x71\xbc\x12\x84\x1f"
    "\xab\xc4\xad\x95\xbf\x76\xfc\x6b\x50\xda\xd3\x05\x13\x4f\x1f\x78"
    "\x81\x83\x6d\xf6\xc5\x6b\x4e\x8b\xc2\x40\x5e\x1d\x0b\x96\xbd\xc0"
    "\x22\xc6\xac\xc1\xd6\x24\x8a\x34\x1a\x5e\xc2\x33\x9a\x3a\x39\xc6"
    "\x06\x08\x22\xd3\x35\xb7\x71\x28\x75\x61\x48\xc1\x64\xe2\x3e\xc7"
    "\x50\x50\xce\xf7\x6c\xd8\x7b\x7c\x01\x84\x55\xb9\xad\x59\x5d\x77"
    "\x5f\x3c\xd4\x6d\x37\xd1\xdd\xf7\xd4\x87\xdb\x8b\x78\x59\x22\x45"
    "\xa3\xbc\xc2\x36\x6a\x2b\x8b\x64\x2a\xc1\xa1\x7c\x2f\x5e\x0a\xbf"
    "\x67\x46\x3f\x8e\xb2\x0d\xa3\xcd\x21\x79\xc5\x81\x8f\xa6\x07\x3a"
    "\x6f\xb3\x65\x44\x75\x3a\x96\xd7\x10\x94\x05\xfa\xb1\x0e\x09\xf2"
    "\xef\xe5\xee\xa1\xd2\xa1\x7c\xb0\xcc\xea\xc8\x8c\xd9\x33\x60\xf2"
    "\x2c\x45\x8b\x80\xe2\xaf\x9e\x10\x67\xab\x9a\xb6\x28\x1a\xad\x1b"
    "\x8e\x48\x62\xaf\xca\xd2\xcc\x70\x76\x59\x16\x12\x4a\x14\x32\xf7"
    "\x77\x76\xd8\x92\xf0\x5b\x34\x3a\x38\xfd\x25\xf5\x8f\xbc\x09\x61"
    "\x2a\xc0\xe5\x2b\x5e\x77\x89\xda\x5d\xb0\x40\xaa\xf4\x5a\x50\xd9"
    "\xe1\xd8\x8c\x07\xa7\x05\xb2\x50\x5c\x71\xb0\x99\xcc\xce\x32\x46"
    "\x62\x18\xf3\xbc\x08\x72\x93\xe7\x38\xd0\xb7\xa4\xc7\xa6\xb0\x62"
    "\x60\x70\x0e\xc9\xa4\x25\xa4\xe3\xe6\xc2\x0f\x42\xcd\x92\x2a\x9e"
    "\xbc\x46\x5c\x5c\x06\xd9\x1e\x5e\xdc\xe3\xc7\xe5\x94\x37\x42\x41"
    "\x62\xdf\xa9\x13\xd4\x7e\x6a\xd6\xc8\x6f\x7d\x34\x8b\x6e\xa0\x65"
    "\xbb\x75\x96\x41\x39\x69\xdd\x3f\x5f\x0d\x74\x8e\xe0\xba\x1f\x73"
    "\x68\x39\x65\x24\x24\x82\xdd\x21\x21\x0e\x51\x93\xb6\xe5\xa4\xaa"
    "\xc9\x75\x15\x3a\xb7\x79\x19\x98\x1d\x0e\xb3\x92\xc7\xce\x99\x7e"
    "\x99\xf5\xee\x6e\xab\x52\x70\xa1\x9c\x43\x57\x8c\x67\x59\x02\x7e"
    "\x4c\x17\x14\x16\x71\x2b\xe8\x74\x9d\x48\xeb\x8d\xd2\xa1\xfd\xfa"
    "\x52\xb2\xf7\x44\xc7\x94\xef\x36\x34\xd5\xb5\xd4\xe0\x92\x62\x86"
    "\x10\x5d\xc7\xf4\xea\x5b\x58\x6f\x71\xae\x5e\x33\x70\x31\x78\x78"
    "\xc5\x06\x18\x27\x6a\xd2\xd4\xa0\xd8\x21\x63\x70\xa2\xc3\xa5\x28"
    "\x1c\x1a\x03\xb0\xab\xca\x88\x4d\xe8\x67\xff\x06\xe6\x37\x8b\x4b"
    "\xb9\x0c\x62\x04\xbe\xa3\xa6\x93\x13\xd3\xbf\x88\x6f\x2e\x2b\x5e"
    "\xfc\x6f\xb1\xb9\x6b\xed\x50\x26\x1b\x21\x32\x84\x1b\x6d\x69\x25"
    "\x53\x9b\x9c\x77\x5a\x72\x39\x57\xaf\xcf\x71\x53\xd9\xb0\x75\x7e"
    "\x6d\x6a\x3c\x31\xc7\x23\x29\x4f\xe9\x94\x04\x31\x64\x22\xee\x2e"
    "\x1d\xde\x73\xd3\x4f\xd9\x29\x98\x3b\x52\xe5\x89\x0a\x52\x41\x15"
    "\x21\x36\x9b\x59\x58\x01\xa6\x1b\xc0\xc0\x09\x28\x2b\xc1\x03\xa6"
    "\xea\x9e\xe3\xe0\xc3\xab\xaa\xad\xe3\x6c\x3c\x2f\xb2\xbf\xb5\x8d"
    "\xe9\x52\x57\xb9\x24\x40\x8d\xa1\x55\x8c\x5d\x9e\x08\xbb\xdd\x15"
    "\x2c\xed\xbb\x6c\xa3\xf9\xec\x02\x6d\x1b\x60\x47\x33\xad\x52\x9a"
    "\xf3\xbd\xdb\x9a\xf1\xb5\x1a\x64\x2b\xfa\x6a\x16\x70\xdb\x91\x2e"
    "\x76\x0d\x0b\x34\x2e\x1f\x37\x11\x9c\x46\xb0\x7e\xd4\x59\x1e\x34"
    "\xc0\x4c\x9b\x0d\xac\xc6\x5a\x8a\x8a\xd8\xd8\xbf\x16\xb1\xd7\xfb"
    "\x50\x1d\x62\x44\x21\x44\x88\xdf\xf0\xbd\x4d\xe7\xc5\x96\xb7\x79"
    "\x62\xf6\x0b\xf5\x34\xd7\x5e\x68\xf1\x21\x1a\x13\x05\x81\x8f\xb9"
    "\x74\x7a\x5c\xbe\x24\x09\x67\xe7\x64\xce\x7e\x7e\x70\xa2\xe8\x11"
    "\x89\x8b\x41\x15\xe3\x51\x57\x0e\x9b\x8f\xdf\x4d\x78\x4e\x2e\xbd"
    "\x37\x67\xea\x4a\xe4\xde\x8e\x5b\x87\xce\xbf\x42\x60\x0c\x45\x18"
    "\x07\x7c\xff\xd0\xb4\x34\x9f\xad\x7a\xf0\xe9\xbf\xec\xe3\xac\xf4"
    "\x59\xf8\x32\xb3\x6f\x37\x7e\x90\x06\x75\x12\xb3\x28\x47\x25\xa1"
    "\x32\x8b\x70\xa6\x3e\x43\xcc\xe0\xe5\x35\xd5\xd6\xb5\xd4\x00\xca"
    "\xbf\x29\xe0\xc1\xeb\x7d\xf1\x4d\x5e\x44\x1c\x1f\x12\xa7\x44\xbd"
    "\x07\xb6\x6c\xd1\x95\x34\x53\xb6\x9a\xee\xca\xf4\x82\x6f\x20\x55"
    "\xa0\x1b\xb3\xe6\x57\x24\x09\x2a\x48\xd7\x67\x88\x2e\x8d\xae\x34"
    "\xae\xae\xef\x67\xd2\xc5\x8a\xc2\x57\x08\x48\x3d\xba\xed\x60\xe4"
    "\xb9\xb7\x38\x59\xee\x9a\x66\x09\x3a\x2a\xcd\x16\xd4\x9f\xcc\x65"
    "\xfd\x01\x48\x8c\x53\xe6\xf3\xd0\xae\x76\xa9\xb6\x6c\x32\x12\xd8"
    "\x5b\x29\xd8\xaf\xe5\xc8\xed\xf6\xa4\x8e\x09\xcf\x4c\x14\x23\x82"
    "\x67\x42\x30\xe7\x50\x21\x88\xe0\x67\x48\xfd\xc2\x00\x42\x0d\x76"
    "\x4e\x01\x52\x5d\x70\x2e\xb5\x07\x39\x8b\x93\xef\x1b\xb5\xaf\x43"
    "\xfc\x46\x26\x5e\xf2\x16\x0c\xb4\x65\xb7\x14\x09\xf9\xbb\xcb\xb4"
    "\x16\xdb\x7d\xf4\x87\x7a\x25\x38\xc6\x14\x4a\xab\xf1\x63\x14\x5c"
    "\x2c\x7e\xe4\xae\xf6\xc0\x57\xea\x1b\x1c\x73\xf2\x3d\xff\x5c\x84"
    "\xb8\x0f\x19\x23\x36\x3c\x29\x6b\x22\x7b\xff\x2d\x18\x24\x81\xfe"
    "\x9a\x54\x3b\xe1\x14\xa2\x1f\xc5\x9a\x64\x58\x49\x88\xe4\x80\x9f"
    "\xe5\x60\x9b\x32\xb8\x42\x0e\x0d\xb9\xc8\x3d\xaa\xb9\x17\xdc\x7c"
    "\x82\x09\x38\x82\x78\xc3\x8e\xd0\x47\xe1\x92\x97\xb6\x46\xb7\xfd"
    "\x09\xe1\x02\x51\x9c\xe2\xeb\xf3\x3e\x06\xce\x43\x20\xbf\x4d\x94"
    "\xc4\x63\x86\x70\xce\x0d\x04\xae\xb7\x48\xeb\xef\xbd\x8f\x5c\x02"
    "\x03\xca\x64\x04\x97\x02\x0a\x01\x12\xe6\xdc\xc3\x8d\xf3\xae\xb3"
    "\xe4\x1d\x65\x89\x71\x50\xf6\xcc\xa4\xd2\xa2\x9f\xdf\x5d\x7f\x0e"
    "\xd8\xab\x07\xb7\x35\xfb\xc3\x27\x5a\xf8\xd0\x23\xb0\x69\x63\xa3"
    "\x1c\x45\x69\x24\x91\xb4\x6f\x45\x64\xee\x81\x98\xf1\x0e\xd5\x2a"
    "\x3e\x21\x8e\x41\xf0\x6a\xc8\x31\x8f\x91\xe5\x27\x89\xad\xf8\x35"
    "\xbb\xd2\xbe\x8e\x6e\xe3\x6b\x0d\x93\x8a\xc4\x0c\x3a\x5d\x77\xda"
    "\x7b\x38\xab\x78\x77\xae\xe4\x5c\x60\x50\xe1\x13\xaf\x91\x7a\x09"
    "\x0d\xb1\x69\xe9\xf5\xb3\x61\xed\xf5\x80\xaf\x8d\x22\x4c\x82\xa4"
    "\x8b\xc4\x07\xac\x30\xea\x1c\xe0\xf6\x15\xd1\x2e\x62\xd4\x10\x47"
    "\xf6\x3f\xa0\xd4\xa2\x7f\x0a\x76\x73\xde\x55\xb6\xc5\x1c\x36\x6e"
    "\x1a\x6f\xb3\xb0\x2c\x0d\xf4\xe8\x6f\x72\x3f\x9b\xb3\xc8\x1e\xb4"
    "\x4f\x5d\x4b\x21\x24\x3c\x82\xfd\x90\xe6\x2f\xdc\x20\x65\xa7\xe9"
    "\x62\x2a\x81\xf3\x48\x9a\x63\xcf\xa0\xa8\x8d\x0b\xab\x82\xda\xc6"
    "\x64\x40\xd4\x28\x51\xf7\xba\x4c\x3c\x10\xc8\x6e\x6f\x0a\x3f\xe8"
    "\xf2\x10\xb7\x13\x7e\xd5\x4f\xe0\x3f\x3d\x46\xb5\x7f\xa7\x51\x7b"
    "\x6b\x7a\xfc\x34\xd9\x4e\x74\x76\x35\x25\x4b\xa4\xb2\x4c\x0c\xdc"
    "\xc7\xed\x7e\x9a\xf4\xe3\xc8\x9c\xe2\x33\x2b\xbe\xf9\x91\xde\x95"
    "\x68\x0f\x0d\x7e\x1e\x43\x80\x3d\x4a\x15\x72\x25\x50\x47\x44\x9e"
    "\xba\xc7\xe4\xcc\x59\x1b\x63\x1a\x16\xc0\x6f\xb2\x08\x19\x96\x1c"
    "\xa5\xbc\x25\xbb\x91\xa4\x1e\x71\x3a\xa3\x3b\xed\x6e\x8e\xd1\xec"
    "\x3b\x74\x94\x59\x02\x3e\xc7\xe6\xef\xed\xb9\x3f\xe5\xd0\xf4\xf9"
    "\x3b\x63\xa1\x8c\x72\x4c\xaf\x90\x68\x1c\x0f\xd0\x4f\x90\x9d\xdb"
    "\x9e\x79\x08\xfb\xe5\xe3\xef\xe7\x64\x8c\xa5\xf1\xe6\x80\xd5\x42"
    "\x91\xd0\x22\xec\xe9\x5b\x32\x6f\xc5\x3c\x21\x5b\x40\x72\x77\x09"
    "\x14\x1e\xf7\x29\x32\x3d\x6d\xcb\x5f\x98\x4a\x47\xcb\x8a\x13\x95"
    "\xa0\xf7\x56\xdf\xc1\x59\x22\x92\x99\x45\xe0\x18\x49\xc0\xe2\x38"
    "\x9f\x65\xb7\x18\xa7\xe3\x60\xef\xf0\xa1\x2f\x46\x24\x2d\x33\xb0"
    "\x56\x9e\x93\x98\x09\x75\xcf\x2d\x8c\x86\x1e\x68\xde\xaf\x9a\x94"
    "\xb4\x20\x6b\x9c\x11\x3a\x92\x93\x08\xed\x69\x04\x02\xff\xed\x2b"
    "\x66\xda\x42\x59\x4b\x41\xc4\x6a\xa4\xf5\x11\x3b\x09\x2d\xdb\x51"
    "\x82\xbe\x5a\x09\x98\x63\xc3\x55\xec\x1f\x70\xe2\x3b\xd8\x2a\x53"
    "\x85\xa3\x4e\x3d\xf4\xae\x56\xb5\x48\x40\x22\x9b\x99\x41\x4f\xd8"
    "\x67\xd1\xeb\xc8\xb8\xec\x0e\xd9\x00\x33\x1a\x44\x9d\x27\xae\x55"
    "\x9a\x2f\x1b\x29\x83\xd7\x67\xf2\x12\x88\x39\xbc\x29\xe4\x66\xac"
    "\x8d\xc3\xe0\x67\xbb\xe0\x77\x4e\xd1\x12\x43\x86\x31\x8c\xf9\x49"
    "\x07\xdc\xb6\x22\x89\xd1\x17\x77\x10\xa3\x6b\xc9\x8a\x69\x3e\xcd"
    "\x7e\x56\x3e\xe1\x5f\x1f\x8e\x58\x58\xc0\xd9\x87\xce\x23\x7f\x03"
    "\xaf\x9d\x81\x8c\x2c\x73\x49\xa8\x26\x4e\x82\x0a\x46\xbe\xfc\x80"
    "\x39\xf2\xec\x28\xff\x0d\x67\x4d\x9f\xd9\xfd\x7f\x8d\x55\x69\x5f"
    "\x5e\x0d\x11\xa2\x8e\x36\x00\x16\xaa\x3a\x58\xc2\x36\x09\xd5\xe2"
    "\x75\x3a\xb1\xc2\xfb\xe5\x4d\x7e\xa1\xcd\x05\xbb\xd8\x92\x67\x45"
    "\x6b\x4d\x79\xec\x6a\x9a\xa7\x00\xeb\x1c\xa2\xfe\xa4\x96\x27\x8c"
    "\x9d\xdc\xc8\xbb\xee\x33\x18\x3b\x81\xd5\x95\x02\x8c\x82\x88\x21"
    "\x28\x56\x12\x98\xb6\x11\xb8\xa5\x64\x1c\xe6\x10\x4a\xd8\xe1\xc4"
    "\xbe\xd0\x35\xb2\x69\x87\x86\x43\xac\xf4\x20\x5a\x0f\xbd\xd4\x31"
    "\x28\x62\x71\xf0\x6d\x08\x24\x7b\xee\xcf\xdf\xdd\x03\x83\x48\xca"
    "\x8a\xb3\xb9\xc3\x04\x0c\xea\xa5\xe6\xaf\x17\x38\x3c\xc5\x31\x33"
    "\x67\xa3\xa3\xd6\xcc\xc9\xe0\xfb\xc2\x90\xc1\x17\xa0\x43\x39\x67"
    "\x78\xc2\x51\x97\x31\x1f\x85\x6e\x83\x49\xa7\xb5\xcf\x7f\x69\x1f"
    "\x99\x97\x44\x48\x08\xfc\x44\xa3\x3a\x88\x62\x72\x2a\x15\x1c\x62"
    "\x60\x9f\x8e\x01\xec\xce\xa4\xed\xe9\xbf\x0f\x52\x93\xa0\x7a\x80"
    "\x63\x4b\x8f\x9e\x12\x3c\x58\x7a\x64\x1a\x72\x17\xa8\x74\x44\xe2"
    "\x8f\x57\xb9\x53\xd0\x00\xfe\x5c\xa2\xa7\xfc\xcf\x89\xe3\xb6\xa6"
    "\x88\x83\x84\x59\xb6\x36\x21\x8e\xf9\xc0\x0b\x34\xc3\x9c\x44\x93"
    "\x15\xd7\x7a\x06\xa4\x28\x84\xcb\xca\x06\x90\x4f\x5c\x5b\x9f\x19"
    "\x09\x4e\xed\xdd\x50\x8b\xab\x62\x05\x24\xeb\x8f\x1a\x30\x42\xad"
    "\xd5\x55\xbf\xac\x52\xe3\xef\x24\x34\xad\x2c\xc2\x35\xb5\xe2\x07"
    "\x1c\xd0\xf5\xe0\xae\x78\xfb\xbb\x84\x0f\xeb\xe3\x8f\x44\xea\xd2"
    "\x33\x75\x1e\x94\xb6\x83\x7d\xe8\x11\xab\x3a\x93\x8c\x81\x6f\x61"
    "\x54\x44\xc9\x29\xc3\x96\x4a\x47\xc8\x88\x5a\xb9\x7a\xd7\x96\x51"
    "\x94\x9b\xab\x4c\x27\xa6\xac\x90\xa6\x4d\xa6\x6a\x14\x8c\xf0\xca"
    "\x84\x59\x2b\xb1\xf6\x65\xb7\xf4\x58\x61\x13\x33\x8b\x10\x2d\x44"
    "\xc0\xca\x5c\xc1\x41\x88\x05\xa7\x65\x2f\xdb\x42\x99\x69\xb0\x7b"
    "\x5c\xcf\x54\xc5\xba\xa4\x10\x15\x18\x0a\x4e\x38\xf9\xec\x02\x56"
    "\xba\xc5\xba\x44\x84\x02\x6a\xcb\xcd\x7e\x0f\x50\x75\xee\x87\x51"
    "\x51\x10\xe6\x18\x8a\xfe\x3b\x8c\x68\x2c\x58\xc9\x34\xa5\x31\xa5"
    "\x3d\x83\x2a\x2e\x27\xc5\xf6\xea\xb8\xa2\x25\xb4\xc8\x9c\x9a\x4e"
    "\xb5\x19\x79\x54\x9b\x3c\x6c\x19\x72\xe4\xa7\x57\xe1\xbf\x42\xed"
    "\x68\x89\x03\xe9\xf9\x12\xf2\x78\xbd\xd0\xd7\xb1\x9b\xf9\xc7\x96"
    "\xbb\xe2\x99\x09\x16\xc8\x00\x0d\x60\x80\x46\xa2\xd5\x5a\xc0\x52"
    "\xc5\xa0\x52\x95\x46\x8e\x56\x3a\x7e\x1a\xbe\xc5\xbd\x70\xf7\x8d"
    "\x15\x6a\x19\x4d\x04\xd0\x2d\xb2\xff\xe1\xe3\xa7\xd3\x6d\x80\x4d"
    "\xa6\xcf\x38\xb8\x06\x6f\x3f\x82\x0f\x52\x1c\x6f\x3e\x5d\x82\xe5"
    "\x00\xe5\xf6\x44\x65\xe2\xa8\x83\x12\xa9\xff\x35\xec\x92\xac\x31"
    "\x84\x41\x81\x9d\xf3\x06\xf5\x2b\x2e\xa6\x16\x92\x63\x65\xc7\x70"
    "\x32\x73\x9b\x80\xea\xbd\x20\xac\xb7\xbb\xf7\x2a\xd7\x2d\xf1\x38"
    "\x36\xe7\x4f\xd5\x0b\x2c\x4c\x32\x21\x9c\xa5\xfd\x97\x06\xb6\xf6"
    "\x40\xbe\xe3\x49\x99\x2c\xd3\xbf\x75\xa2\x47\x18\x9d\x37\xa7\x0f"
    "\xd3\xe5\x1d\xc1\x05\xb4\xa0\xa1\xa1\xe6\xfb\x39\x0f\x2f\xa8\x81"
    "\x00\x3e\xcf\xef\x98\x85\x0e\xe2\x62\x00\x78\x66\xc3\xcb\xc0\xbb"
    "\x41\x35\xfd\xa9\xaf\x7e\x98\x94\x60\xa5\xca\x75\xd3\x28\xba\x6d"
    "\xc6\x3c\xb5\x17\x01\xfb\x50\xbe\x9f\xa8\xec\xb0\x41\x16\x97\x3a"
    "\x0d\xdf\x10\xc7\x1f\x2f\x24\xf4\xf3\x2e\xbe\xb3\x5f\x87\x1f\xb2"
    "\xc9\xe8\x06\xb4\x6d\x77\x9d\x3a\x6b\xa1\xad\x12\x87\xf2\x1b\x00"
    "\x09\x15\xda\x52\xa4\x21\xc9\x43\x1d\x03\xe4\x01\xa6\x93\x57\xb5"
    "\x27\x04\xbf\x4a\x39\x84\xc9\xed\x0f\x1f\xf2\xee\xe9\x51\xff\xe6"
    "\xa5\x88\xb1\x0d\xcc\xad\xeb\x33\xfc\x23\xf0\x6b\xe5\xa4\xf5\xcc"
    "\x88\xbc\x76\x5c\x16\x45\x2a\x61\x4f\x14\x9c\xf5\x07\x0c\xe4\x3d"
    "\x38\xc3\x32\x15\x63\xe5\x70\xd2\x87\x22\x0f\x5f\xf3\xa1\xc7\xcd"
    "\xcf\xaf\xaa\x2e\xa1\xed\x86\x98\xcf\x6a\x87\x5f\xdd\xbc\x53\x7d"
    "\x93\x38\x59\x0d\xda\x97\x4c\x01\x84\x5b\xe2\x8b\x4f\x90\x12\x86"
    "\x3b\xf8\x7f\x2d\xbd\x62\x04\x2e\xc1\x93\xc0\x97\x94\x40\xaf\x88"
    "\x47\x3c\xb3\xec\x65\x5c\x5a\x8b\x69\x03\x1f\xc0\x29\x05\xd6\x6a"
    "\x7e\xef\xe3\xe0\xf9\xcd\x0d\x88\x3e\x0b\xee\x9a\x84\x7c\xe4\x45"
    "\xef\x0a\x20\x1b\x27\x63\x2b\x8c\xda\x32\x9a\x17\xad\xc9\xf2\xa8"
    "\x33\xd5\x85\x15\x4a\x71\x65\x79\xa2\xa8\xe3\xed\x69\x38\x10\x54"
    "\x98\x62\xa7\xfc\x93\x68\x6a\x18\xd5\x55\x23\xe1\xf5\x86\x0a\xf9"
    "\xc9\x97\x17\x35\xab\x9e\x2a\x79\x85\xc6\x70\xf4\x0c\xa4\x80\xa9"
    "\x33\x36\x33\xfb\x02\x34\xf2\x6b\x96\xa8\x73\x2a\xa3\x7d\x4c\xb7"
    "\x02\xcf\x58\x15\x2d\x53\x1d\x9a\xcf\x96\xb3\x5e\x05\xce\x29\xb9"
    "\xd7\x7d\xff\xa4\x9c\x56\x07\xe7\x6a\x48\xea\x81\xfb\x2f\x15\x5e"
    "\x8e\xde\x81\x77\x1d\x13\x6a\x9d\xef\xb7\xc5\x7c\x6c\x1a\x73\x16"
    "\x59\x8a\x6f\x24\x85\xf4\xe6\x32\xfa\x9e\xc2\xea\xc2\xcb\xcf\xe5"
    "\xbe\x93\xb8\x52\xc1\xf1\x6c\x05\xd1\x70\x34\xdd\x7b\xe3\x7f\xcf"
    "\x8a\x30\xa8\x1d\xf2\xf9\x25\x09\xa2\x4d\x05\x4b\xba\xa8\x63\x2a"
    "\xb1\xf3\x87\x34\xd8\xdf\x8c\x13\xd8\x72\xb5\x20\x36\x1d\xaf\x7c"
    "\x2a\x8f\xaa\xb4\xd1\xfc\x0b\xc9\xf2\x21\xa2\x07\x8a\xfe\x40\xf0"
    "\x52\x36\x3d\x32\x82\x67\x09\x5d\xf5\x59\xe5\xd2\x39\x09\x91\x49"
    "\xce\x11\x9f\x53\xe1\xbb\x32\x6a\xc5\xf4\x74\x60\xdd\x19\x9e\xf1"
    "\x96\x82\x42\x28\x74\x51\x95\x21\x07\xc0\x8f\xc4\x54\x5f\x12\x66"
    "\x58\x7c\xfe\x24\xb6\xea\x11\xe0\x76\x96\xb1\x25\x45\xa4\xcc\xbf"
    "\x71\x09\x91\xdd\x1e\x47\x8d\x9b\xe8\xcf\xe5\xd3\x19\x87\xaa\x63"
    "\x11\xae\xa5\x50\xa7\xeb\x1f\x6d\x47\x4e\xb8\x7a\x5b\x63\x15\x0f"
    "\x85\x7b\x65\xb1\xd2\x30\xad\x25\x0a\x9c\xb8\xeb\x7a\x62\xb9\x9b"
    "\xb9\x41\x1d\x00\xdd\xb6\x4b\xa0\xa1\xad\xb9\x26\xbd\x7f\xf0\x97"
    "\xd6\x87\xa8\x9c\x57\x40\x9b\xe6\x78\xad\x6f\xbb\xa4\xad\x3e\xb1"
    "\xdd\x3b\x16\x56\x19\xbd\xa5\x31\xf6\xd7\xb5\x2f\xdc\x72\x7f\x2d"
    "\xcb\xff\x02\x09\x8d\x3d\xbb\xcd\x32\x98\xf3\x2a\x0d\x5b\xdc\xd9"
    "\xa4\x99\x59\xab\x70\xa3\xa2\xec\x40\x79\x01\x57\xaa\x38\xb7\xa5"
    "\x76\x92\x92\x1a\x61\x22\xca\x03\x99\x4b\x51\xe0\x01\x6a\xca\x57"
    "\x25\xa6\x56\x4a\x87\xcd\xe8\x9d\x6a\x72\xbd\x2c\x0d\x3c\x30\x5b"
    "\x2c\x52\x60\xbd\x55\xb0\xac\x21\x6a\x98\xda\x5d\xaa\xb6\x57\x91"
    "\xcf\xf0\x81\x3c\xff\xca\x48\x4d\x3a\x94\x19\xfe\xaf\xc1\x55\xba"
    "\x6e\x80\x4d\x49\x24\x29\x81\xbc\x92\xdc\xb2\x89\xa4\x70\x84\x8a"
    "\x2c\xb3\x77\xe3\x73\x10\x47\x51\x0f\x13\x32\x0b\xba\xb5\x70\x88"
    "\x62\x63\x37\x38\xb6\xbb\x4f\xfb\x36\x52\xa8\xc1\xa1\xf0\x8e\x2c"
    "\xad\xce\xf3\x0a\xb0\xfb\xad\x90\x07\x26\xc2\xc1\xeb\x02\xa7\x6d"
    "\xd7\x54\x52\xbe\x0c\x28\xb8\xdf\x13\xd4\xfc\x4a\x37\x36\x8f\x62"
    "\x5a\xde\x5f\x43\xa3\x89\xde\x72\x79\x2d\x5a\x87\x14\x14\x81\x3f"
    "\x07\x2c\xf0\x20\x6a\xf8\xf2\xf2\xb9\x55\x0e\x9a\xda\x03\xa4\x53"
    "\x37\x64\x9c\xa2\xc8\x87\x51\x4a\x25\x10\x9a\x11\x15\xbd\xe6\xea"
    "\x24\xa1\xd9\xfa\x95\x02\x56\x6c\x9b\xe6\x5d\x3f\xdc\xca\x69\x4b"
    "\x3e\x51\xba\x7a\xf1\x71\x92\xd9\x25\x84\x80\x40\x3d\xd2\x15\xf0"
    "\xe5\x65\x3a\xe4\x43\x7a\xd0\x33\xab\xb8\x39\x25\xe0\xf1\x9a\xbf"
    "\x9e\x67\xf5\x73\x74\x70\xab\x98\x9f\x90\xeb\xf4\xfd\x5c\xcd\xb4"
    "\x81\xab\x9a\x85\xf1\x8a\x56\xaa\x10\x30\xd2\x9a\xee\x1b\x60\x1d"
    "\xd1\xfe\x70\xfb\xa6\x02\x94\xbf\x64\x4a\x2e\xad\xa1\x83\x06\x5c"
    "\x71\x00\x88\x28\x85\x14\xbb\x67\x13\xf4\xd7\x15\x5c\x69\xfe\xc6"
    "\x2b\xbe\xdb\xf3\x94\xe4\x95\xeb\x42\xf0\x8c\x21\x25\xd0\x29\x6d"
    "\x24\xc3\x7a\xae\x23\xac\x29\x8d\x15\x03\x36\x87\x3d\x2c\x0d\x98"
    "\x75\xd5\x96\x3c\x41\xc1\x88\x3b\xca\x25\x80\x41\xdc\x2f\x19\x6c"
    "\x9e\x68\x76\x0f\x18\x9d\x2e\x69\xda\x8c\x08\xd2\x7d\x92\x7e\xf9"
    "\x18\xf9\x7e\x2c\xba\xc9\xd6\xcd\x66\x6f\x52\xa6\x06\xca\xd6\x01"
    "\xec\x52\x02\xd1\x97\x2b\xba\xaf\x01\xd0\x90\xe8\xb0\x20\xe9\x53"
    "\xe7\xcb\x37\x35\x4d\x25\xf1\x0f\xb6\x14\x54\x48\x55\xe9\x0d\xfa"
    "\x2e\x66\x39\x07\x18\x3d\xb1\xe4\x87\x29\x64\xb1\x7c\x0c\x47\xf0"
    "\x46\xa9\x8e\xef\x03\x74\x26\x0e\x9b\x5f\x5a\x6f\x97\x21\x55\x35"
    "\x80\x8c\x34\x9c\x9f\xb3\x14\x30\xa8\xb7\xc6\x40\x0d\x4f\xe5\x5b"
    "\x6e\xc9\xe7\xb5\x41\x78\x21\x16\x76\x67\x89\xfe\xd1\xd4\x4b\x5f"
    "\x84\x4d\x96\xc7\x82\xa6\x0d\x65\xc7\x9c\x31\xf1\xc6\x3e\x87\xff"
    "\xc9\xe0\x2c\xf9\x77\x95\x08\x3f\xbd\x00\xfc\x61\xeb\x77\x21\xe6"
    "\x35\x1d\xdb\xd4\xe6\x8b\x6f\xbb\x82\x80\x5d\x58\x9d\x1d\x7a\x59"
    "\x80\x64\xc6\xe2\x66\x7e\x1a\x4a\x4c\x71\xf2\xd8\xe2\xff\x02\x7f"
    "\xc2\x1d\x32\x33\xc6\xe1\xf4\x05\x90\xd1\x55\x1a\x15\x65\xd2\x73"
    "\xdc\x04\x40\x81\xd2\x40\xe3\x59\x35\x9f\x9c\x3c\xdd\x72\x4f\x2d"
    "\xfd\xa8\x9e\x1c\x85\x1c\xac\x3d\xb9\xa0\x2d\x10\xc2\x57\x12\x65"
    "\x51\xda\x80\x03\x34\x52\xd2\xff\xdd\x18\x0f\xbd\x02\x6b\x93\x07"
    "\x64\x88\x13\x2a\x80\x8e\xe5\x8a\xec\x5b\xe9\xae\xf3\x94\x34\x48"
    "\xa5\x88\xb1\x0d\xcc\xad\xeb\x33\xfc\x23\xf0\x6b\xe5\xa4\xf5\xcc"
    "\x88\xbc\x76\x5c\x16\x45\x2a\x61\x4f\x14\x9c\xf5\x07\x0c\xe4\x3d"
    "\x38\xc3\x32\x15\x63\xe5\x70\xd2\x87\x22\x0f\x5f\xf3\xa1\xc7\xcd"
    "\xcf\xaf\xaa\x2e\xa1\xed\x86\x98\xcf\x6a\x87\x5f\xdd\xbc\x53\x7d"
    "\x93\x38\x59\x0d\xda\x97\x4c\x01\x84\x5b\xe2\x8b\x4f\x90\x12\x86"
    "\x3b\xf8\x7f\x2d\xbd\x62\x04\x2e\xc1\x93\xc0\x97\x94\x40\xaf\x88"
    "\x47\x3c\xb3\xec\x65\x5c\x5a\x8b\x69\x03\x1f\xc0\x29\x05\xd6\x6a"
    "\x7e\xef\xe3\xe0\xf9\xcd\x0d\x88\x3e\x0b\xee\x9a\x84\x7c\xe4\x45"
    "\xef\x0a\x20\x1b\x27\x63\x2b\x8c\xda\x32\x9a\x17\xad\xc9\xf2\xa8"
    "\x8d\x5e\x97\x32\x66\x4f\xab\x7a\xc3\x55\xf1\x6c\xb6\xf4\x3f\xc2"
    "\xf9\xd5\xbe\xfd\x07\x83\x8d\x0f\x9f\xf5\xa0\x88\x13\x71\x53\xc8"
    "\x37\xa5\x6d\xb5\x84\xce\xee\xb3\x03\x9b\xdc\x86\x24\xe9\x41\x68"
    "\xa5\x44\xab\x59\x69\xa9\x85\xeb\x60\x8a\x9b\x73\x4a\x46\x84\x68"
    "\x70\x36\xbb\x77\x0d\xfc\xa0\x03\xc1\xbc\xa8\xdd\xde\xaa\x53\xab"
    "\xd0\x34\x15\x6e\xcf\x1c\x57\x2d\xa8\xaf\x14\xf0\x0d\x07\x2d\x76"
    "\x03\xfd\xbc\x41\xf4\x18\x26\xe1\x11\x32\x79\x57\x0c\xa5\x55\xd4"
    "\x20\x3b\xa0\x7b\x76\xa9\x62\xe0\x6f\x5b\xb8\xf0\x59\x06\x25\x4a"
    "\xdd\x7c\x15\x08\x1f\x69\x0a\xa2\x20\x84\x9b\x58\x97\x58\x70\x61"
    "\x66\x1d\x79\x00\xc8\xad\x08\x50\xbf\xeb\xd8\x12\xb2\xf6\xbb\x06"
    "\xac\xa8\x76\x49\xc5\x69\xae\xc3\x99\x0e\x4d\xb0\x27\x90\x5f\x8c"
    "\x2e\xab\xe3\x3b\xb5\xfc\x36\x51\x83\x71\x2f\x1c\x54\x95\xa2\xef"
    "\x13\xf1\xe4\x71\x66\x4c\xce\x81\x95\x98\x96\x10\x3a\xde\xe1\x55"
    "\xaa\x2d\x65\x3f\xf8\xf3\xe2\x93\x6a\xc7\x6e\xba\x4d\xcd\xbd\xc1"
    "\x62\xf4\xb3\x6b\x70\xdd\xf2\x21\x61\x9e\xb4\x10\x52\x2e\xef\xa9"
    "\x13\x97\x47\x86\x61\x9a\x25\xa7\x1c\xc8\xc0\x52\xfb\x75\x18\xda"
    "\x2a\x4e\x5d\xfc\xc2\x63\xfd\x5c\x2a\x54\x30\x42\x51\xce\x64\x14"
    "\xd6\xe1\xae\x61\xd0\x06\xdb\xe1\x51\x6d\x03\x9b\x2f\x34\xf7\x87"
    "\x91\x0e\x0b\xca\xab\x1c\x91\xdc\x2a\xb2\x48\x93\x38\x61\x8f\x7f"
    "\xef\x2f\x05\xcc\xc1\xc5\x2c\x3c\x35\xb1\x9d\xc3\x05\x73\xe8\xb4"
    "\x0e\x73\x3d\x0d\x8b\x53\x09\x03\x27\xb5\x8d\xfe\x5f\xe8\x87\x46"
    "\x46\xfa\xcb\x5f\x1a\xe7\x6b\xda\x8a\xd7\x59\x4f\x0c\x45\xad\x68"
    "\x44\x4c\xd8\x0b\x75\x3f\x7e\x47\x6c\xe2\x8d\xcc\xb6\x5b\x54\x84"
    "\x64\x22\xb7\x46\x67\x14\xfd\x77\xbe\xec\xc1\x5d\x06\x9c\x16\x26"
    "\x09\x57\x99\x6a\x01\x7e\xac\x51\xb1\x61\x44\x5c\xf1\x11\xf9\x6e"
    "\x46\xd4\xea\x36\xcf\x7b\xf8\x2c\x13\x6b\xba\xd4\x20\x3b\x81\x59"
    "\x4d\x4d\x92\xf3\x24\x95\x0a\x75\x55\x50\x70\x58\x74\x04\x78\x5a"
    "\xd0\x14\xda\xc3\xd9\x38\x10\xfd\x70\x86\xc2\xa0\x6d\x98\x69\xdc"
    "\x53\x0a\x5a\x2f\x92\xcf\x4a\x21\x3c\x2d\xcd\xa9\xc7\x8e\x17\xa7"
    "\x5c\x1f\xb3\x06\x6a\xa4\xad\xb7\x47\x92\x98\x25\xb3\x7b\x25\xef"
    "\x7f\x7f\x45\x56\x8c\xf4\x55\xe1\x17\x09\x34\xf9\x8c\x69\x06\x91"
    "\x5c\xde\xca\xcb\xae\xa4\x0b\xd7\x53\x4f\xb5\x74\x53\xd2\xf1\xfb"
    "\xd4\xa2\x3c\xda\xf1\xd5\xda\x77\xbb\x33\x75\x06\x8a\xf6\x41\xe2"
    "\x20\xff\x59\xe3\x9c\xf2\x74\xa7\x90\x2d\xa2\x01\x47\x5c\x8a\x58"
    "\x83\xfe\xdd\x86\xa9\x7c\x86\xf5\xf3\xd5\x3a\x3f\xcc\x69\x72\x76"
    "\xc6\x84\x5b\xec\x6b\xc4\xb1\xed\x50\x8c\x5c\x09\x12\x6e\x92\x0d"
    "\xb9\x65\xf1\x02\x34\x27\x2b\xe8\x88\x2e\x89\x0e\xdb\xac\x17\x52"
    "\x5a\x36\x1d\xe7\xea\xff\x72\xeb\x1f\xd4\x78\x7d\xc4\xeb\x59\x32"
    "\x32\x9f\x09\xa8\xd2\x19\xfa\x11\x7f\xf8\x68\x09\x46\x68\x57\x77"
    "\x64\x86\x1e\x5c\xef\x44\xae\x05\x33\x32\x83\xd1\xbc\xce\x3f\x5f"
    "\xff\x16\x6d\xfb\x44\xea\xb6\xfc\xf3\xff\x22\xc3\x61\x58\xe0\xb8"
    "\x64\xd0\x48\xba\x17\x93\x85\xd5\xa5\x3a\x22\x96\xe1\xc3\x55\x6f"
    "\x6e\xf4\xc9\x5d\x01\x7f\xe0\xd6\x92\xd7\x8d\x03\x30\x3d\x44\x7a"
    "\x6c\xec\x48\xb6\xf0\xdc\xac\xa9\x88\x6f\x4b\x10\xdf\x31\x47\x86"
    "\xe8\xf3\xf1\xa5\x92\x73\xa2\xb6\xbf\x55\x6c\xc5\xbc\xb9\xf0\x05"
    "\xab\xfe\xf0\x76\xe0\xc8\x96\x79\xd0\x07\x05\xfc\x52\xe3\xef\x89"
    "\x8f\x22\xbe\xee\x94\x50\x28\xb0\xf1\x9b\xc2\x7a\xc2\x28\x0f\xd8"
    "\xd7\xa9\xbb\x8b\x99\x6e\xd5\x68\x8c\x9b\x0f\x08\xaf\xf2\x20\xc1"
    "\xf0\x2b\xd4\x1f\x4c\x26\x04\xa5\x55\x8d\xb9\x9d\x42\x1b\x0e\xf9"
    "\xca\xc3\x9a\x6e\x63\x24\x97\xde\x8b\xdf\x92\x9e\xa9\x29\x85\x91"
    "\xd7\xd0\x8b\x42\x68\xca\x86\xe7\x65\x58\x02\x66\x3d\x62\x2b\x38"
    "\x2a\x3c\x93\x70\xdc\xfd\xdf\x3e\xd7\x78\x83\x43\x48\x9d\x9c\x3e"
    "\x8f\xf7\x17\x9c\xe1\xe2\x06\x81\x5e\xad\xa9\xd5\x4c\x68\xed\xc9"
    "\xf2\x75\xcb\x27\x07\x00\x50\x96\xa6\xa9\xff\xad\x74\xda\xb1\xab"
    "\x41\xbb\xd8\xba\xb0\xcc\xf7\x3f\xc3\x10\xdb\xa7\x63\xf6\x26\x01"
    "\x79\x54\x08\x73\x7c\x3f\xae\xec\x65\xbb\x10\x87\x71\x16\x06\x2a"
    "\xf6\x1d\xf6\x67\x1b\x68\x6a\x97\x57\x4f\x46\x94\x08\x34\x68\x53"
    "\x1a\xfd\x5a\xb8\x7b\xa9\x4a\xfa\x5e\x42\xe4\x74\x20\x6a\xb7\x3f"
    "\x27\x3c\xc2\x04\xa5\x6e\x89\x9e\x13\x7a\x3f\x62\xf8\xfe\xa1\x57"
    "\x9c\x0d\x43\x25\x29\x69\x88\x98\x23\x48\x4e\x8a\x4a\x11\x61\x88"
    "\xa6\x2c\x97\x8a\x75\x8d\x8f\x6e\x29\xe7\x4b\x5d\x4e\x80\xc0\xff"
    "\xc3\xbd\xe1\x68\xb6\x1e\x13\xab\xd9\x72\x4b\x32\x7d\x76\xa3\xa6"
    "\xa7\x0b\xe3\xaf\x3c\x7d\x35\x83\x46\x2c\x99\x86\xe5\xee\x57\xa7"
    "\x7e\x14\x3d\x42\xb7\x1f\x6f\x22\x0e\x61\x6f\x92\x7d\x73\xef\xd0"
    "\x29\x53\x64\x50\xf0\xfc\x49\x8f\x1a\xc2\xaf\x44\x0b\x99\xc8\x7d"
    "\x55\xb0\xf3\xaf\x18\xc8\x18\x00\x29\x67\x27\xee\x84\x59\x87\xf3"
    "\x69\x1d\x0b\xcc\x08\x78\xe0\xe4\xb7\xb1\x5f\x29\xc5\x4f\xaa\x9d"
    "\x37\xe6\x79\x6d\xfb\x57\xe2\xc3\xd3\x7a\xe9\x26\x92\x25\xc8\x8b"
    "\x97\x98\x0d\x9b\xa2\x64\x85\x37\xa2\x32\x14\xfe\x28\xa7\x61\x06"
    "\xd8\xe6\x1a\x80\xa1\xcf\x97\xad\xcd\xe1\x34\xdb\xa4\x4b\xec\x40"
    "\x2f\xa2\xab\xca\xdb\xd5\x0a\xed\x7d\xe2\x72\x24\x9f\x76\xf5\x02"
    "\xda\xfa\xec\x2d\x39\x10\xa0\xed\x2e\x12\xa9\x40\x62\xa4\xfd\x3c"
    "\x85\x93\x98\x9d\x90\x3b\xcb\xe1\x1f\xf9\xcf\x3b\x8a\x30\x4a\x8e"
    "\xdd\xb1\x26\x40\x04\x71\x64\x7b\xf4\xa8\x10\xce\x05\xbf\x99\x93"
    "\x68\x0d\x2b\x56\x51\xe8\xbb\x9f\x96\x98\x65\x3f\x7b\xa0\x04\x27"
    "\xcb\xbe\x75\xaf\x5c\x7d\x35\x78\x6c\x52\xb3\x43\x38\xed\x8a\xa0"
    "\xea\x28\x12\xf5\x19\xd2\x04\x1f\xc1\xb0\xff\xbc\x96\x52\x8c\x4c"
    "\x58\x88\x91\x1b\x3d\x06\xf6\x83\x43\x02\x71\x38\xba\x1b\xcd\x6f"
    "\x7c\xf2\x09\xb1\xc5\x71\x21\x3b\xe5\xe6\x04\xae\xf9\x7c\xef\x8c"
    "\xa8\xf6\x9b\xc1\x06\xdc\x85\x31\x7a\x58\x40\x1a\x32\x9e\xd5\xaf"
    "\x9f\x96\x3f\x37\x73\x15\xc9\x48\x48\xf9\x53\x5d\x80\x46\xa4\x03"
    "\x67\xd1\xeb\xc8\xb8\xec\x0e\xd9\x00\x33\x1a\x44\x9d\x27\xae\x55"
    "\x9a\x2f\x1b\x29\x83\xd7\x67\xf2\x12\x88\x39\xbc\x29\xe4\x66\xac"
    "\x8d\xc3\xe0\x67\xbb\xe0\x77\x4e\xd1\x12\x43\x86\x31\x8c\xf9\x49"
    "\x07\xdc\xb6\x22\x89\xd1\x17\x77\x10\xa3\x6b\xc9\x8a\x69\x3e\xcd"
    "\x7e\x56\x3e\xe1\x5f\x1f\x8e\x58\x58\xc0\xd9\x87\xce\x23\x7f\x03"
    "\xaf\x9d\x81\x8c\x2c\x73\x49\xa8\x26\x4e\x82\x0a\x46\xbe\xfc\x80"
    "\x39\xf2\xec\x28\xff\x0d\x67\x4d\x9f\xd9\xfd\x7f\x8d\x55\x69\x5f"
    "\x5e\x0d\x11\xa2\x8e\x36\x00\x16\xaa\x3a\x58\xc2\x36\x09\xd5\xe2"
    "\x75\x3a\xb1\xc2\xfb\xe5\x4d\x7e\xa1\xcd\x05\xbb\xd8\x92\x67\x45"
    "\x6b\x4d\x79\xec\x6a\x9a\xa7\x00\xeb\x1c\xa2\xfe\xa4\x96\x27\x8c"
    "\x9d\xdc\xc8\xbb\xee\x33\x18\x3b\x81\xd5\x95\x02\x8c\x82\x88\x21"
    "\x28\x56\x12\x98\xb6\x11\xb8\xa5\x64\x1c\xe6\x10\x4a\xd8\xe1\xc4"
    "\xbe\xd0\x35\xb2\x69\x87\x86\x43\xac\xf4\x20\x5a\x0f\xbd\xd4\x31"
    "\x28\x62\x71\xf0\x6d\x08\x24\x7b\xee\xcf\xdf\xdd\x03\x83\x48\xca"
    "\x8a\xb3\xb9\xc3\x04\x0c\xea\xa5\xe6\xaf\x17\x38\x3c\xc5\x31\x33"
    "\x67\xa3\xa3\xd6\xcc\xc9\xe0\xfb\xc2\x90\xc1\x17\xa0\x43\x39\x67"
    "\x78\xc2\x51\x97\x31\x1f\x85\x6e\x83\x49\xa7\xb5\xcf\x7f\x69\x1f"
    "\x99\x97\x44\x48\x08\xfc\x44\xa3\x3a\x88\x62\x72\x2a\x15\x1c\x62"
    "\x60\x9f\x8e\x01\xec\xce\xa4\xed\xe9\xbf\x0f\x52\x93\xa0\x7a\x80"
    "\x63\x4b\x8f\x9e\x12\x3c\x58\x7a\x64\x1a\x72\x17\xa8\x74\x44\xe2"
    "\x8f\x57\xb9\x53\xd0\x00\xfe\x5c\xa2\xa7\xfc\xcf\x89\xe3\xb6\xa6"
    "\x88\x83\x84\x59\xb6\x36\x21\x8e\xf9\xc0\x0b\x34\xc3\x9c\x44\x93"
    "\x15\xd7\x7a\x06\xa4\x28\x84\xcb\xca\x06\x90\x4f\x5c\x5b\x9f\x19"
    "\x09\x4e\xed\xdd\x50\x8b\xab\x62\x05\x24\xeb\x8f\x1a\x30\x42\xad"
    "\xd5\x55\xbf\xac\x52\xe3\xef\x24\x34\xad\x2c\xc2\x35\xb5\xe2\x07"
    "\x1c\xd0\xf5\xe0\xae\x78\xfb\xbb\x84\x0f\xeb\xe3\x8f\x44\xea\xd2"
    "\x33\x75\x1e\x94\xb6\x83\x7d\xe8\x11\xab\x3a\x93\x8c\x81\x6f\x61"
    "\xeb\xdb\x91\x1d\xe5\x98\x05\x17\x13\xcb\xef\x67\xfc\x4a\xbf\x5f"
    "\x1d\xb4\xed\xe9\xc5\x39\xe9\xd6\x77\x93\x99\x39\xc3\x30\x16\x08"
    "\x56\x54\xe8\xb8\xd9\x1e\xa4\x04\x1e\x56\xb2\x2c\xef\x4e\xd4\xa1"
    "\x4d\xba\xf1\xad\x23\x12\x45\x9c\xf6\xae\x84\xc3\xfe\x49\xd5\xba"
    "\xc8\xc3\xde\xf8\x20\xfc\xf0\x1b\x84\xb0\x5c\x9e\x27\x48\x46\x0a"
    "\x94\x59\xfe\x8f\x3b\xe0\x5d\x81\xb3\xc9\x47\xe8\xa1\x59\x0d\x27"
    "\x11\x0d\x04\xec\xf0\x0a\x62\x3b\xbe\xbd\x99\x9d\x86\x6e\xfc\xe0"
    "\x19\x7b\xf1\x8f\xc4\x64\xdd\x86\x51\x6d\xcf\x3b\x1e\x52\xa7\x30"
    "\xd1\x8d\xe0\xed\x3f\x4c\x27\x64\x9c\x07\x37\xdb\x03\xb3\x47\x62"
    "\x03\x95\x19\x9c\x35\x13\x34\xf5\x09\xc0\x99\x97\xd2\x52\x59\xbe"
    "\x3c\xb5\xb7\xba\x4b\x63\xe5\x0c\xf1\x19\x5a\x16\xed\x63\x3e\x15"
    "\xbd\x3f\xc5\x6f\x27\xed\x7d\x4f\x22\x0a\x43\x59\xab\x31\x3c\xdf"
    "\x8c\xb9\x77\x84\xab\x35\xdb\x92\xd6\xaa\xc0\x3c\xf0\x9f\xc4\x02"
    "\x5a\xbf\x10\x75\x5d\xda\xe9\xe3\x55\xc9\x63\x4a\xfc\xee\x5e\xeb"
    "\x6e\x05\x41\x02\xa9\x61\xbd\x4b\x08\x8c\x4d\x98\x9a\x84\x25\x13"
    "\x98\xd2\x85\x90\x24\xcc\x39\x30\x4b\x06\x8c\xf0\x28\x8d\xf9\x50"
    "\x56\x3f\x8d\xa3\x2e\xd1\x3e\x63\x08\x21\xad\xb0\x66\x78\xa0\xeb"
    "\x29\xf9\xc7\x13\x4e\xe7\x0c\x6f\xaf\xd9\x05\x76\x7c\xcf\x59\x84"
    "\x19\x3d\x5c\x35\x92\xa7\xe2\x2e\xfb\x90\x1a\xbd\xf8\x26\x7d\xcb"
    "\xba\xdb\x8b\x98\xb8\x6b\xd0\x95\xc2\x4e\x41\xe1\x7e\x68\xfd\x4c"
    "\xa2\x06\x91\x65\x55\x29\xfd\x29\x80\x90\xfa\xd6\xf4\xc1\x28\x80"
    "\x68\xfc\xe3\x23\xf2\x0a\xca\x71\x35\x16\xfa\xf9\x64\x3a\x50\x79"
    "\x1b\x20\x38\x05\x56\xf2\x3a\x96\x99\x46\xb4\x2f\xbf\x1c\x98\x69"
    "\x4d\x66\x55\xdc\x70\xbf\x0b\xca\x51\x91\x20\x09\xc5\x51\xbb\x1b"
    "\x4e\x9d\xab\x65\x1b\x13\x79\x8d\x2a\x37\x5a\x68\xdc\xdc\xa2\x96"
    "\x10\x98\xee\x08\x91\x87\x6f\x52\x63\x6b\x0e\x33\x0f\xeb\xfb\x90"
    "\x95\x4e\x5f\x86\xb9\x22\x40\x6f\x38\x56\xb7\x57\x65\x52\xad\x0d"
    "\xa9\xc9\x8f\x48\xe4\x44\xd2\xc0\x92\x80\x8a\x19\x7e\x04\x60\x2b"
    "\x04\x3b\xd2\x05\x18\xef\x0b\xd5\xb9\x79\xed\x17\x6b\xf8\x4c\x1b"
    "\xae\x01\x2e\xa6\x2b\xf1\x93\x9b\x05\xa2\x94\xbe\x81\x80\x24\x3f"
    "\xf1\x28\xd6\xdf\xe8\x48\xaa\x25\xab\x25\xbb\x55\x99\xfc\x4f\x82"
    "\x8f\xfd\x5e\x40\xe7\x52\x9c\xd6\xc1\xd0\x3d\x46\xdb\xde\xe7\x12"
    "\x0b\x10\xea\x88\xd8\x00\xa0\xcc\xee\x4b\x54\x6b\x77\x9d\x5d\xad"
    "\xee\x1d\x68\x9d\x68\xe3\xc9\x73\xd4\x14\x66\x0e\x82\x1d\x98\x1b"
    "\x4d\x6c\xe2\x53\xf9\xcc\x1c\x16\x87\x34\xc8\xe2\x60\x72\x58\x04"
    "\x36\x29\xdb\xe8\x59\x3c\x43\x7f\x7a\xca\xfa\x96\x14\x51\x11\xeb"
    "\x86\xc7\x3f\x7e\x53\x6f\x32\x45\x14\x8d\xb4\x4f\xcb\xf4\xb2\x26"
    "\x83\xc3\x3d\x59\x2d\x22\x3f\x14\x9a\x62\x67\xcf\xc7\xa8\xde\xbe"
    "\x91\x6d\xc9\x4d\xb1\xe6\xa0\x76\x4c\xda\x67\x09\xd8\xd5\xec\x22"
    "\xe5\x0b\x99\x20\x5c\x7c\xd9\x73\x40\xf3\x4d\x32\x90\xbd\xe1\x9a"
    "\x08\xd2\x3d\x9a\x15\xb0\xb9\x25\x56\x47\x63\x60\xea\xe8\x1c\xfb"
    "\x92\xe8\xfb\x2c\x2e\x34\xdb\x1d\xe5\x0e\x34\x05\xc4\x82\x4e\x5f"
    "\xeb\xfd\x85\xf1\xda\x9d\x07\xf5\x80\x6d\xeb\x42\xee\xbc\x9d\x71"
    "\xbd\xb7\x5a\xb8\x07\x56\x30\xb2\x49\x5d\xae\xa4\x3a\xf8\x70\x2a"
    "\xdc\x54\xbb\x56\x00\x68\x10\x53\xc2\x9b\x91\xfa\xf6\x34\x6e\xcc"
    "\x4b\xce\x96\xed\x56\xac\x10\x84\x89\x3d\x41\x99\x84\xda\xc5\xa7"
    "\x9f\x0f\xd7\x28\x3a\xad\x51\x31\x9e\x15\xe0\xd4\x5d\x95\x11\xc1"
    "\x3f\x8b\x54\xf5\x80\x81\x5d\x1e\xe9\xe4\xa0\x3e\xa3\xa7\x1f\xf4"
    "\x28\xc1\x7b\x20\xba\x08\xc6\x8c\x01\xdb\x48\xde\x28\x58\x5c\x60"
    "\xa2\xc6\x6b\x72\x8b\x35\xae\xc8\xa7\x84\xa4\x8b\x1c\x10\x3e\x58"
    "\x6f\x1b\xd9\x9d\xe8\x53\x10\x64\x79\xce\x3f\xf7\xae\x4c\x66\x9c"
    "\xd2\x9a\x4f\x25\xe5\x77\xca\x05\xe4\x98\x3d\xc0\x47\x53\x66\x56"
    "\x0b\xc0\x39\x35\x83\xbd\x0c\x83\x47\xb7\x4b\xf6\xaa\xa7\x42\x10"
    "\x58\xb8\x40\x22\xec\x8c\x0f\x03\x60\x8e\x9f\x43\x1e\xff\x63\xd0"
    "\x4b\xea\x42\x5b\x20\xaf\x91\x23\x5d\x78\xa3\x17\x8c\x10\x8d\x7c"
    "\xbc\x25\xe1\x08\xd7\x96\x0a\x1f\x96\x9a\x03\x0e\x50\xa8\xc3\xed"
    "\x0a\xe5\x9d\x57\x96\xd3\xc6\xfc\xe2\x32\xeb\x7a\xf0\xb4\xf3\x12"
    "\xc8\x74\x74\x6c\x48\xcc\xa6\xa6\x7a\xa5\xa8\x67\xfb\xcb\xc5\xb0"
    "\x5d\x69\x79\x88\xa3\x37\xba\x93\x90\x54\x3d\x37\x74\xd4\x61\xfc"
    "\x0a\xc5\x50\x9f\xab\x75\x6f\x63\xcc\x5b\x84\x80\x65\xff\x1c\xb4"
    "\x8f\x58\x63\xc0\x4f\x90\x53\x88\x2c\x60\x79\x6d\x89\xd1\x40\xd2"
    "\x06\x89\x8d\xda\xc1\x12\x90\x3b\xb4\xf9\x03\x0d\xc5\x2c\x85\x73"
    "\x6e\x7a\xec\x22\x3a\x99\x9f\x1b\xb5\x44\x44\x58\x2c\x03\x41\xfc"
    "\xbd\x9a\xc5\xe9\xc6\x53\xc6\x8d\x8a\x1d\x8e\xf1\x66\xb9\x1e\xd6"
    "\xe4\x83\xc2\x0c\x72\xc8\x96\x6e\x2d\xba\xb2\xec\xd3\x4d\xaf\x86"
    "\xf9\x51\xd7\xc3\xca\x0b\x8c\x58\xc6\x3c\x7b\x6e\xab\x93\x7f\x41"
    "\xf4\x33\x3b\x10\xc7\xdb\x90\x77\x84\x91\x6b\x03\x21\x21\x3c\x28"
    "\xf4\x29\xfa\x47\xe5\x5f\x4b\xef\xc8\xa8\xa2\xb0\x0a\x16\xaf\x5f"
    "\xed\xe8\x36\xc9\xf0\x06\x01\x64\xce\x1f\x72\x9e\xa8\xd9\xd1\x51"
    "\x42\x16\x0d\x85\x46\xbe\x92\x86\xd7\x6e\xcc\xce\x49\xbb\x55\x0f"
    "\x53\x1b\xc8\x5f\x37\xe5\x50\x85\x6c\x56\x89\xc0\x41\x65\x37\x3e"
    "\x9a\x79\x2c\x9e\x69\x47\x4a\x35\x3f\xab\x98\xb1\x86\xe5\x19\x3c"
    "\xa2\x2f\x2d\xc5\x51\x07\x1a\xcd\xa2\xfe\xbf\xeb\xed\x5b\x09\xfc"
    "\xff\x2f\xe3\xb6\x6b\xbe\xab\x19\x22\x4f\xea\xe4\xf4\x4e\x69\xeb"
    "\x56\x8e\x06\x5a\x25\xa9\xe6\x7b\x8d\x9d\xf9\xf4\x5b\x54\x22\xce"
    "\x26\x81\x23\x84\x63\x09\x3f\x4a\x8f\x61\x89\x37\x4d\x86\x5d\x78"
    "\x0c\x40\x7f\x36\x85\xa0\x6c\x66\x68\xcb\x30\x2f\x0f\x12\xa7\x40"
    "\x49\xd1\xe9\x4d\x46\x75\x70\xf1\xe7\xfd\x4b\x93\x2a\x95\xfa\x1c"
    "\x4c\x20\x26\x42\x42\x8d\xb5\x90\x37\x1b\xf7\x2a\x29\x5c\xe4\xdb"
    "\x05\xe0\x03\x5d\x89\x86\x9d\xdb\xc8\xb9\xd1\x94\xae\x40\x01\x76"
    "\x4c\xd2\xe0\x16\xe9\xdc\x26\xbb\xd9\x89\x44\x8e\x1f\x03\xf6\xd7"
    "\x66\xef\x87\x8a\xc1\x75\x5d\xf7\xa2\x18\xe9\x38\xc2\x2a\x75\x94"
    "\x07\xa6\x31\x7d\xcb\xd3\x17\xbe\x89\xc9\x81\x78\x39\x90\xde\x6d"
    "\x03\x14\x01\xe3\x2d\xf9\x1b\xc0\x19\x7f\x3d\xe3\xf7\x5d\xad\x68"
    "\x7f\x69\x79\x4d\x59\x92\x44\xd3\xd3\x12\xfd\x86\x80\x98\xd0\x4e"
    "\x8a\x33\x4c\x17\x3d\xe4\xa1\x67\x8a\x4d\x2d\xeb\x0c\xb6\x49\x06"
    "\xef\x24\x57\xba\xe7\x0e\xfa\x05\xd1\xd3\x68\xa0\xa1\x3a\xb6\xa9"
    "\x0d\xb7\x38\x05\x47\x8f\x4c\x20\xb2\xc1\x8c\x0a\x3a\x5c\x1b\xf8"
    "\x1b\x27\x00\xa4\xb6\x76\x2a\x09\x21\x34\xc1\x30\x16\x0a\x8e\xf4"
    "\x12\x8b\x32\x9d\x39\x93\xda\x57\x7e\xdb\x53\x24\x8d\x63\x82\x86"
    "\x5b\x94\x8d\x3a\xa6\xb9\x3b\x2b\x49\xeb\x61\x71\xbc\x12\x84\x1f"
    "\xab\xc4\xad\x95\xbf\x76\xfc\x6b\x50\xda\xd3\x05\x13\x4f\x1f\x78"
    "\x81\x83\x6d\xf6\xc5\x6b\x4e\x8b\xc2\x40\x5e\x1d\x0b\x96\xbd\xc0"
    "\x22\xc6\xac\xc1\xd6\x24\x8a\x34\x1a\x5e\xc2\x33\x9a\x3a\x39\xc6"
    "\x06\x08\x22\xd3\x35\xb7\x71\x28\x75\x61\x48\xc1\x64\xe2\x3e\xc7"
    "\x50\x50\xce\xf7\x6c\xd8\x7b\x7c\x01\x84\x55\xb9\xad\x59\x5d\x77"
    "\x5f\x3c\xd4\x6d\x37\xd1\xdd\xf7\xd4\x87\xdb\x8b\x78\x59\x22\x45"
    "\xa3\xbc\xc2\x36\x6a\x2b\x8b\x64\x2a\xc1\xa1\x7c\x2f\x5e\x0a\xbf"
    "\x67\x46\x3f\x8e\xb2\x0d\xa3\xcd\x21\x79\xc5\x81\x8f\xa6\x07\x3a"
    "\x6f\xb3\x65\x44\x75\x3a\x96\xd7\x10\x94\x05\xfa\xb1\x0e\x09\xf2"
    "\xef\xe5\xee\xa1\xd2\xa1\x7c\xb0\xcc\xea\xc8\x8c\xd9\x33\x60\xf2"
    "\x2c\x45\x8b\x80\xe2\xaf\x9e\x10\x67\xab\x9a\xb6\x28\x1a\xad\x1b"
    "\x8e\x48\x62\xaf\xca\xd2\xcc\x70\x76\x59\x16\x12\x4a\x14\x32\xf7"
    "\x77\x76\xd8\x92\xf0\x5b\x34\x3a\x38\xfd\x25\xf5\x8f\xbc\x09\x61"
    "\x2a\xc0\xe5\x2b\x5e\x77\x89\xda\x5d\xb0\x40\xaa\xf4\x5a\x50\xd9"
    "\xe1\xd8\x8c\x07\xa7\x05\xb2\x50\x5c\x71\xb0\x99\xcc\xce\x32\x46"
    "\x62\x18\xf3\xbc\x08\x72\x93\xe7\x38\xd0\xb7\xa4\xc7\xa6\xb0\x62"
    "\x60\x70\x0e\xc9\xa4\x25\xa4\xe3\xe6\xc2\x0f\x42\xcd\x92\x2a\x9e"
    "\xbc\x46\x5c\x5c\x06\xd9\x1e\x5e\xdc\xe3\xc7\xe5\x94\x37\x42\x41"
    "\x62\xdf\xa9\x13\xd4\x7e\x6a\xd6\xc8\x6f\x7d\x34\x8b\x6e\xa0\x65"
    "\xbb\x75\x96\x41\x39\x69\xdd\x3f\x5f\x0d\x74\x8e\xe0\xba\x1f\x73"
    "\x68\x39\x65\x24\x24\x82\xdd\x21\x21\x0e\x51\x93\xb6\xe5\xa4\xaa"
    "\xc9\x75\x15\x3a\xb7\x79\x19\x98\x1d\x0e\xb3\x92\xc7\xce\x99\x7e"
    "\x99\xf5\xee\x6e\xab\x52\x70\xa1\x9c\x43\x57\x8c\x67\x59\x02\x7e"
    "\x4c\x17\x14\x16\x71\x2b\xe8\x74\x9d\x48\xeb\x8d\xd2\xa1\xfd\xfa"
    "\x52\xb2\xf7\x44\xc7\x94\xef\x36\x34\xd5\xb5\xd4\xe0\x92\x62\x86"
    "\x10\x5d\xc7\xf4\xea\x5b\x58\x6f\x71\xae\x5e\x33\x70\x31\x78\x78"
    "\xc5\x06\x18\x27\x6a\xd2\xd4\xa0\xd8\x21\x63\x70\xa2\xc3\xa5\x28"
    "\x1c\x1a\x03\xb0\xab\xca\x88\x4d\xe8\x67\xff\x06\xe6\x37\x8b\x4b"
    "\xb9\x0c\x62\x04\xbe\xa3\xa6\x93\x13\xd3\xbf\x88\x6f\x2e\x2b\x5e"
    "\xfc\x6f\xb1\xb9\x6b\xed\x50\x26\x1b\x21\x32\x84\x1b\x6d\x69\x25"
    "\x53\x9b\x9c\x77\x5a\x72\x39\x57\xaf\xcf\x71\x53\xd9\xb0\x75\x7e"
    "\x6d\x6a\x3c\x31\xc7\x23\x29\x4f\xe9\x94\x04\x31\x64\x22\xee\x2e"
    "\x1d\xde\x73\xd3\x4f\xd9\x29\x98\x3b\x52\xe5\x89\x0a\x52\x41\x15"
    "\x21\x36\x9b\x59\x58\x01\xa6\x1b\xc0\xc0\x09\x28\x2b\xc1\x03\xa6"
    "\xea\x9e\xe3\xe0\xc3\xab\xaa\xad\xe3\x6c\x3c\x2f\xb2\xbf\xb5\x8d"
    "\xe9\x52\x57\xb9\x24\x40\x8d\xa1\x55\x8c\x5d\x9e\x08\xbb\xdd\x15"
    "\x2c\xed\xbb\x6c\xa3\xf9\xec\x02\x6d\x1b\x60\x47\x33\xad\x52\x9a"
    "\xf3\xbd\xdb\x9a\xf1\xb5\x1a\x64\x2b\xfa\x6a\x16\x70\xdb\x91\x2e"
    "\x76\x0d\x0b\x34\x2e\x1f\x37\x11\x9c\x46\xb0\x7e\xd4\x59\x1e\x34"
    "\xc0\x4c\x9b\x0d\xac\xc6\x5a\x8a\x8a\xd8\xd8\xbf\x16\xb1\xd7\xfb"
    "\x50\x1d\x62\x44\x21\x44\x88\xdf\xf0\xbd\x4d\xe7\xc5\x96\xb7\x79"
    "\x62\xf6\x0b\xf5\x34\xd7\x5e\x68\xf1\x21\x1a\x13\x05\x81\x8f\xb9"
    "\x74\x7a\x5c\xbe\x24\x09\x67\xe7\x64\xce\x7e\x7e\x70\xa2\xe8\x11"
    "\x89\x8b\x41\x15\xe3\x51\x57\x0e\x9b\x8f\xdf\x4d\x78\x4e\x2e\xbd"
    "\x37\x67\xea\x4a\xe4\xde\x8e\x5b\x87\xce\xbf\x42\x60\x0c\x45\x18"
    "\x07\x7c\xff\xd0\xb4\x34\x9f\xad\x7a\xf0\xe9\xbf\xec\xe3\xac\xf4"
    "\x59\xf8\x32\xb3\x6f\x37\x7e\x90\x06\x75\x12\xb3\x28\x47\x25\xa1"
    "\x32\x8b\x70\xa6\x3e\x43\xcc\xe0\xe5\x35\xd5\xd6\xb5\xd4\x00\xca"
    "\xbf\x29\xe0\xc1\xeb\x7d\xf1\x4d\x5e\x44\x1c\x1f\x12\xa7\x44\xbd"
    "\x07\xb6\x6c\xd1\x95\x34\x53\xb6\x9a\xee\xca\xf4\x82\x6f\x20\x55"
    "\xa0\x1b\xb3\xe6\x57\x24\x09\x2a\x48\xd7\x67\x88\x2e\x8d\xae\x34"
    "\xae\xae\xef\x67\xd2\xc5\x8a\xc2\x57\x08\x48\x3d\xba\xed\x60\xe4"
    "\xb9\xb7\x38\x59\xee\x9a\x66\x09\x3a\x2a\xcd\x16\xd4\x9f\xcc\x65"
    "\xfd\x01\x48\x8c\x53\xe6\xf3\xd0\xae\x76\xa9\xb6\x6c\x32\x12\xd8"
    "\x5b\x29\xd8\xaf\xe5\xc8\xed\xf6\xa4\x8e\x09\xcf\x4c\x14\x23\x82"
    "\x67\x42\x30\xe7\x50\x21\x88\xe0\x67\x48\xfd\xc2\x00\x42\x0d\x76"
    "\x4e\x01\x52\x5d\x70\x2e\xb5\x07\x39\x8b\x93\xef\x1b\xb5\xaf\x43"
    "\xfc\x46\x26\x5e\xf2\x16\x0c\xb4\x65\xb7\x14\x09\xf9\xbb\xcb\xb4"
    "\x16\xdb\x7d\xf4\x87\x7a\x25\x38\xc6\x14\x4a\xab\xf1\x63\x14\x5c"
    "\x2c\x7e\xe4\xae\xf6\xc0\x57\xea\x1b\x1c\x73\xf2\x3d\xff\x5c\x84"
    "\xb8\x0f\x19\x23\x36\x3c\x29\x6b\x22\x7b\xff\x2d\x18\x24\x81\xfe"
    "\x9a\x54\x3b\xe1\x14\xa2\x1f\xc5\x9a\x64\x58\x49\x88\xe4\x80\x9f"
    "\xe5\x60\x9b\x32\xb8\x42\x0e\x0d\xb9\xc8\x3d\xaa\xb9\x17\xdc\x7c"
    "\x82\x09\x38\x82\x78\xc3\x8e\xd0\x47\xe1\x92\x97\xb6\x46\xb7\xfd"
    "\x09\xe1\x02\x51\x9c\xe2\xeb\xf3\x3e\x06\xce\x43\x20\xbf\x4d\x94"
    "\xc4\x63\x86\x70\xce\x0d\x04\xae\xb7\x48\xeb\xef\xbd\x8f\x5c\x02"
    "\x03\xca\x64\x04\x97\x02\x0a\x01\x12\xe6\xdc\xc3\x8d\xf3\xae\xb3"
    "\xe4\x1d\x65\x89\x71\x50\xf6\xcc\xa4\xd2\xa2\x9f\xdf\x5d\x7f\x0e"
    "\xd8\xab\x07\xb7\x35\xfb\xc3\x27\x5a\xf8\xd0\x23\xb0\x69\x63\xa3"
    "\x1c\x45\x69\x24\x91\xb4\x6f\x45\x64\xee\x81\x98\xf1\x0e\xd5\x2a"
    "\x3e\x21\x8e\x41\xf0\x6a\xc8\x31\x8f\x91\xe5\x27\x89\xad\xf8\x35"
    "\xbb\xd2\xbe\x8e\x6e\xe3\x6b\x0d\x93\x8a\xc4\x0c\x3a\x5d\x77\xda"
    "\x7b\x38\xab\x78\x77\xae\xe4\x5c\x60\x50\xe1\x13\xaf\x91\x7a\x09"
    "\x0d\xb1\x69\xe9\xf5\xb3\x61\xed\xf5\x80\xaf\x8d\x22\x4c\x82\xa4"
    "\x8b\xc4\x07\xac\x30\xea\x1c\xe0\xf6\x15\xd1\x2e\x62\xd4\x10\x47"
    "\xf6\x3f\xa0\xd4\xa2\x7f\x0a\x76\x73\xde\x55\xb6\xc5\x1c\x36\x6e"
    "\x1a\x6f\xb3\xb0\x2c\x0d\xf4\xe8\x6f\x72\x3f\x9b\xb3\xc8\x1e\xb4"
    "\x4f\x5d\x4b\x21\x24\x3c\x82\xfd\x90\xe6\x2f\xdc\x20\x65\xa7\xe9"
    "\x62\x2a\x81\xf3\x48\x9a\x63\xcf\xa0\xa8\x8d\x0b\xab\x82\xda\xc6"
    "\x64\x40\xd4\x28\x51\xf7\xba\x4c\x3c\x10\xc8\x6e\x6f\x0a\x3f\xe8"
    "\xf2\x10\xb7\x13\x7e\xd5\x4f\xe0\x3f\x3d\x46\xb5\x7f\xa7\x51\x7b"
    "\x6b\x7a\xfc\x34\xd9\x4e\x74\x76\x35\x25\x4b\xa4\xb2\x4c\x0c\xdc"
    "\xc7\xed\x7e\x9a\xf4\xe3\xc8\x9c\xe2\x33\x2b\xbe\xf9\x91\xde\x95"
    "\x68\x0f\x0d\x7e\x1e\x43\x80\x3d\x4a\x15\x72\x25\x50\x47\x44\x9e"
    "\xba\xc7\xe4\xcc\x59\x1b\x63\x1a\x16\xc0\x6f\xb2\x08\x19\x96\x1c"
    "\xa5\xbc\x25\xbb\x91\xa4\x1e\x71\x3a\xa3\x3b\xed\x6e\x8e\xd1\xec"
    "\x3b\x74\x94\x59\x02\x3e\xc7\xe6\xef\xed\xb9\x3f\xe5\xd0\xf4\xf9"
    "\x3b\x63\xa1\x8c\x72\x4c\xaf\x90\x68\x1c\x0f\xd0\x4f\x90\x9d\xdb"
    "\x9e\x79\x08\xfb\xe5\xe3\xef\xe7\x64\x8c\xa5\xf1\xe6\x80\xd5\x42"
    "\x91\xd0\x22\xec\xe9\x5b\x32\x6f\xc5\x3c\x21\x5b\x40\x72\x77\x09"
    "\x14\x1e\xf7\x29\x32\x3d\x6d\xcb\x5f\x98\x4a\x47\xcb\x8a\x13\x95"
    "\xa0\xf7\x56\xdf\xc1\x59\x22\x92\x99\x45\xe0\x18\x49\xc0\xe2\x38"
    "\x9f\x65\xb7\x18\xa7\xe3\x60\xef\xf0\xa1\x2f\x46\x24\x2d\x33\xb0"
    "\x56\x9e\x93\x98\x09\x75\xcf\x2d\x8c\x86\x1e\x68\xde\xaf\x9a\x94"
    "\xb4\x20\x6b\x9c\x11\x3a\x92\x93\x08\xed\x69\x04\x02\xff\xed\x2b"
    "\x66\xda\x42\x59\x4b\x41\xc4\x6a\xa4\xf5\x11\x3b\x09\x2d\xdb\x51"
    "\x82\xbe\x5a\x09\x98\x63\xc3\x55\xec\x1f\x70\xe2\x3b\xd8\x2a\x53"
    "\x85\xa3\x4e\x3d\xf4\xae\x56\xb5\x48\x40\x22\x9b\x99\x41\x4f\xd8"
    "\x67\xd1\xeb\xc8\xb8\xec\x0e\xd9\x00\x33\x1a\x44\x9d\x27\xae\x55"
    "\x9a\x2f\x1b\x29\x83\xd7\x67\xf2\x12\x88\x39\xbc\x29\xe4\x66\xac"
    "\x8d\xc3\xe0\x67\xbb\xe0\x77\x4e\xd1\x12\x43\x86\x31\x8c\xf9\x49"
    "\x07\xdc\xb6\x22\x89\xd1\x17\x77\x10\xa3\x6b\xc9\x8a\x69\x3e\xcd"
    "\x7e\x56\x3e\xe1\x5f\x1f\x8e\x58\x58\xc0\xd9\x87\xce\x23\x7f\x03"
    "\xaf\x9d\x81\x8c\x2c\x73\x49\xa8\x26\x4e\x82\x0a\x46\xbe\xfc\x80"
    "\x39\xf2\xec\x28\xff\x0d\x67\x4d\x9f\xd9\xfd\x7f\x8d\x55\x69\x5f"
    "\x5e\x0d\x11\xa2\x8e\x36\x00\x16\xaa\x3a\x58\xc2\x36\x09\xd5\xe2"
    "\x75\x3a\xb1\xc2\xfb\xe5\x4d\x7e\xa1\xcd\x05\xbb\xd8\x92\x67\x45"
    "\x6b\x4d\x79\xec\x6a\x9a\xa7\x00\xeb\x1c\xa2\xfe\xa4\x96\x27\x8c"
    "\x9d\xdc\xc8\xbb\xee\x33\x18\x3b\x81\xd5\x95\x02\x8c\x82\x88\x21"
    "\x28\x56\x12\x98\xb6\x11\xb8\xa5\x64\x1c\xe6\x10\x4a\xd8\xe1\xc4"
    "\xbe\xd0\x35\xb2\x69\x87\x86\x43\xac\xf4\x20\x5a\x0f\xbd\xd4\x31"
    "\x28\x62\x71\xf0\x6d\x08\x24\x7b\xee\xcf\xdf\xdd\x03\x83\x48\xca"
    "\x8a\xb3\xb9\xc3\x04\x0c\xea\xa5\xe6\xaf\x17\x38\x3c\xc5\x31\x33"
    "\x67\xa3\xa3\xd6\xcc\xc9\xe0\xfb\xc2\x90\xc1\x17\xa0\x43\x39\x67"
    "\x78\xc2\x51\x97\x31\x1f\x85\x6e\x83\x49\xa7\xb5\xcf\x7f\x69\x1f"
    "\x99\x97\x44\x48\x08\xfc\x44\xa3\x3a\x88\x62\x72\x2a\x15\x1c\x62"
    "\x60\x9f\x8e\x01\xec\xce\xa4\xed\xe9\xbf\x0f\x52\x93\xa0\x7a\x80"
    "\x63\x4b\x8f\x9e\x12\x3c\x58\x7a\x64\x1a\x72\x17\xa8\x74\x44\xe2"
    "\x8f\x57\xb9\x53\xd0\x00\xfe\x5c\xa2\xa7\xfc\xcf\x89\xe3\xb6\xa6"
    "\x88\x83\x84\x59\xb6\x36\x21\x8e\xf9\xc0\x0b\x34\xc3\x9c\x44\x93"
    "\x15\xd7\x7a\x06\xa4\x28\x84\xcb\xca\x06\x90\x4f\x5c\x5b\x9f\x19"
    "\x09\x4e\xed\xdd\x50\x8b\xab\x62\x05\x24\xeb\x8f\x1a\x30\x42\xad"
    "\xd5\x55\xbf\xac\x52\xe3\xef\x24\x34\xad\x2c\xc2\x35\xb5\xe2\x07"
    "\x1c\xd0\xf5\xe0\xae\x78\xfb\xbb\x84\x0f\xeb\xe3\x8f\x44\xea\xd2"
    "\x33\x75\x1e\x94\xb6\x83\x7d\xe8\x11\xab\x3a\x93\x8c\x81\x6f\x61"
    "\xc0\xf8\x27\xa9\x6a\x46\x13\xf2\x98\x60\x99\x04\xd9\x7a\xb8\x99"
    "\x8c\x3c\x79\x64\x99\x0b\x55\x84\x6b\xe0\x3f\xd8\x93\xff\x17\x39"
    "\x6e\x1b\x82\x9b\x66\xb3\x80\x0a\xcb\xe6\x17\x75\x8f\x62\x53\x97"
    "\x21\x14\xe5\xc4\x63\x9a\xe0\xb0\x9e\xfe\x5a\x9c\x3a\xa0\xcd\xa7"
    "\x3a\x80\x28\x08\x99\x8a\x87\x3f\x67\x47\x11\xab\x78\xd1\x97\x4b"
    "\x1e\xb3\x90\x20\x0a\x78\xf5\x3d\xa4\xe0\x0d\x31\xbd\x3b\x4e\x20"
    "\x7f\x4e\x33\x76\xed\xd7\x6c\x37\xa3\xb2\x76\x23\x85\x41\xb3\x8a"
    "\x19\xb4\x8a\x7a\xcb\x04\x5a\x8e\x74\xe9\xf0\x2f\xe8\x2c\xb2\x73"
    "\xe1\x4b\xe0\x42\x79\xd4\xf9\x67\xe7\x79\xca\xc7\x79\xfc\x20\x76"
    "\x34\xb9\xe4\x00\x48\x2c\xc5\x56\x61\xe2\x81\x35\x39\xed\xfb\xa1"
    "\x62\xaf\x56\x12\xe0\x57\xed\x10\xc7\x48\xf3\x27\x2f\x47\xbd\x53"
    "\xf9\x37\xe3\xf9\x4e\x73\x29\xa1\xe3\xa3\x53\x62\x95\xab\xcd\xb7"
    "\xc9\xe2\x17\xe9\x46\xfd\xfa\x34\x3b\xd5\xac\x26\x72\x44\x7f\x0c"
    "\xcf\x2b\x69\xdc\xae\x93\xfe\x7f\xc6\x73\x1e\x26\xf3\xf0\xe8\xc4"
    "\xa6\xa4\x6c\x74\x07\xd2\x83\x58\xa9\x29\x58\xe1\xb6\x08\xf6\x67"
    "\x2a\x42\xcc\xeb\x34\xae\x2c\x3b\x7d\x1f\xcc\x39\x15\xc5\x93\x3c",

    "1818e87564e0c50974ecaabbb2eb4ca2f6cc820234b51861e2590be625f1f703",

    "310fbb2027bdb7042f0e09e7b092e9ada506649510a7aa029825c8e8019e9c30"
    "749d723f2de1bd8c043d8d89d3748c2f",

    "a001636f3ff1ce34f432f8e8f7785b78be84318beb8485a406650a8b243c419f"
    "7db6435cf6bf3000c6524adb5b52bad01afb76b3ceff701331e18b85b0e4cbd3"
  }
};

/*
 * String comparison that returns the position in the string where the
 * first difference is encountered:
 */
int posstrcmp(char *a, char *b) {
	int	i = 0;

	while (*a - *b == 0) {
		if (*a == 0) {
			/* Strings match */
			return 0;
		}
		a++;
		b++;
		i++;
	}
	/* Strings don't match */
	return i + 1;
}

/* 128 spaces: */
const static char *spacer =
    "                                                                "
    "                                                                ";

int main(int argc, char **argv) {
	int	i, errors, t;
	char	md[SHA512_DIGEST_STRING_LENGTH];

	errors = 0;
	for (i = 0; i <  NUM_TEST_VECTORS; i++) {
		printf("SHA-256 TEST #%d...", i+1);
		SHA256_Data(test_vectors[i].input_data, test_vectors[i].data_len, md);
		if ((t = posstrcmp(md, test_vectors[i].sha256_result)) != 0) {
			printf("ERROR:\n");
			printf("  Expected: '0x%s'\n", test_vectors[i].sha256_result);
			printf("  Output:   '0x%s'\n", md);
			printf("               %s^--Differ at position %d\n", &spacer[128-t+1], t);
			errors++;
		} else {
			printf("OK\n");
		}

		printf("SHA-384 TEST #%d...", i+1);
		SHA384_Data(test_vectors[i].input_data, test_vectors[i].data_len, md);
		if ((t = posstrcmp(md, test_vectors[i].sha384_result)) != 0) {
			printf("ERROR:\n");
			printf("  Expected: '0x%s'\n", test_vectors[i].sha384_result);
			printf("  Output:   '0x%s'\n", md);
			printf("               %s^--Differ at position %d\n", &spacer[128-t+1], t);
			errors++;
		} else {
			printf("OK\n");
		}

		printf("SHA-512 TEST #%d...", i+1);
		SHA512_Data(test_vectors[i].input_data, test_vectors[i].data_len, md);
		if ((t = posstrcmp(md, test_vectors[i].sha512_result)) != 0) {
			printf("ERROR:\n");
			printf("  Expected: '0x%s'\n", test_vectors[i].sha512_result);
			printf("  Output:   '0x%s'\n", md);
			printf("               %s^--Differ at position %d\n", &spacer[128-t+1], t);
			errors++;
		} else {
			printf("OK\n");
		}
	}
	if (errors) {
		printf("TOTAL ERRORS: %d\n", errors);
	} else {
		printf("ALL TEST VECTORS PASSED!\n");
	}

	return 1;
}

